import multiprocessing as mp
import os
from concurrent.futures import ProcessPoolExecutor
from typing import List, Optional, Tuple

import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from beetroots.inversion.plots.plots_estimator import PlotsEstimator
from beetroots.inversion.results.abstract_results import ResultsExtractor
from beetroots.inversion.results.utils.perf_saver import EstimatorPerfSaver
from beetroots.space_transform.abstract_transform import Scaler


class ResultsExtractorOptimMLE(ResultsExtractor):
    """This class does not use the same utils as optim_MAP and MCMC to exploit the pixel-wise nature of the optimisation approach.

    .. warning::

        Unfinished class
    """

    def __init__(
        self,
        path_data_csv_out_mle: str,
        path_img: str,
        path_raw: str,
        N_MCMC: int,
        T_OPTI_MLE: int,
        freq_save: int,
        max_workers: int,
        N: int,
        D: int,
    ):
        r"""

        Parameters
        ----------
        path_data_csv_out_mle : str
            path to the csv file in which the performance of estimators is to be saved
        path_img : str
            path to the folder in which images are to be saved
        path_raw : str
            path to the raw ``.hdf5`` files
        N_MCMC : int
            number of optimization procedures to run per posterior distribution
        T_OPTI_MLE : int
            total size of each optimization procedure
        freq_save : int
            frequency of saved iterates, 1 means that all iterates were saved (used to show correct optimization procedure sizes in chain plots)
        max_workers : int
            maximum number of workers that can be used for results extraction
        N : int
            number of pixels / components :math:`n`, i.e., number of inverse problems that were solved
        D : int
            dimension of each vector of physical parameter :math:`\theta_n`
        """
        self.path_data_csv_out_mle = path_data_csv_out_mle
        r"""str: path to the csv file in which the performance of estimators is to be saved"""
        self.path_img = path_img
        r"""str: path to the folder in which images are to be saved"""
        self.path_raw = path_raw
        r"""str: path to the raw ``.hdf5`` files"""

        self.N_MCMC = N_MCMC
        r"""int: number of optimization procedures to run per posterior distribution"""
        self.T_OPTI_MLE = T_OPTI_MLE
        r"""int: total size of each optimization procedure"""
        self.freq_save = freq_save
        r"""int: frequency of saved iterates, 1 means that all iterates were saved (used to show correct optimization procedure sizes in chain plots)"""
        self.effective_T_MLE = self.T_OPTI_MLE // freq_save
        r"""int: effective length of the optimization procedures in ``.hdf5`` files"""

        self.N = N
        r"""number of pixels / components :math:`n`, i.e., number of inverse problems that were solved"""
        self.D = D
        r"""dimension of each vector of physical parameter :math:`\theta_n`"""

        self.max_workers = max_workers
        r"""int: maximum number of workers that can be used for results extraction"""

    @classmethod
    def read_estimator(
        cls,
        path_data_csv_out_mle: str,
        model_name: str,
    ) -> Tuple[np.ndarray, pd.DataFrame]:
        """reads the value of an already estimated MLE from a csv file.

        Parameters
        ----------
        path_data_csv_out_optim_mle : str
            path to the csv file containing an already estimated MLE
        model_name : str
            name of the model, used to identify the posterior distribution

        Returns
        -------
        np.ndarray
            MLE estimator
        pd.DataFrame
            original DataFrame read from the csv file
        """
        path_file = f"{path_data_csv_out_mle}/results_MLE.csv"
        assert os.path.isfile(path_file), "The MLE has not been computed yet."

        df_results_mle = pd.read_csv(path_file)

        df_mle_best = (
            df_results_mle[df_results_mle["model_name"] == model_name]
            .groupby("n")["objective"]
            .min()
        )
        df_mle_best = df_mle_best.reset_index()

        df_mle_final = pd.merge(
            df_results_mle, df_mle_best, on=["n", "objective"], how="inner"
        )
        df_mle_final = df_mle_final.sort_values("n")
        df_mle_final = df_mle_final.drop_duplicates(["n", "objective"])

        N = df_mle_final["n"].max()
        D = df_mle_final["d"].max()

        Theta_MLE = np.zeros((N, D))
        for d in range(D):
            Theta_MLE[:, d] = df_mle_final.loc[:, f"x_MLE_{d}_lin"].values

        return Theta_MLE, df_mle_final

    def extract_mle_results(
        self,
        list_model_names: List[str],
        scaler: Scaler,
        Theta_true_scaled: Optional[np.ndarray] = None,
    ):
        global _extract_mle_results_1_pix

        if Theta_true_scaled is not None:
            Theta_true_lin = scaler.from_scaled_to_lin(Theta_true_scaled)

        def _extract_mle_results_1_pix(dict_input: dict) -> dict:
            n = dict_input["n"]
            seed = dict_input["seed"]
            model_name = dict_input["model_name"]

            path_results = f"{self.path_raw}/{model_name}/"
            path_results += f"opti_MLE_{seed}/pixel_{n}"

            with h5py.File(f"{path_results}/mc_chains.hdf5", "r") as f:
                list_Theta_lin = np.array(f["list_Theta"])  # (T, 1, D)
                list_objective = np.array(f["list_objective"])  # (T,)

            perf_saver = EstimatorPerfSaver()
            x_MLE_lin, objective = perf_saver.estimate_point_with_lowest_obj(
                list_Theta_lin,
                list_objective,
            )
            x_MLE_scaled = scaler.from_linear_to_scaled(
                x_MLE_lin.reshape((1, self.D)),
            )

            if Theta_true_scaled is not None:
                x_n_scaled_true = Theta_true_scaled[n, :].reshape((1, self.D))
                mse = perf_saver.compute_MSE(x_MLE_scaled, x_n_scaled_true)
                snr = perf_saver.compute_SNR(x_MLE_scaled, x_n_scaled_true)
            else:
                mse = None
                snr = None

            dict_output = {
                "n": n,
                "seed": seed,
                "model_name": model_name,
                "MSE": mse,
                "SNR": snr,
                "objective": objective,
            }
            for d in range(self.D):
                dict_output[f"x_MLE_{d}_lin"] = x_MLE_lin[0, d]
                if Theta_true_scaled is not None:
                    dict_output[f"Theta_true_{d}_lin"] = Theta_true_lin[n, d]

            return dict_output

        # * global function
        list_params = [
            {"n": n, "seed": seed, "model_name": model_name}
            for n in range(self.N)
            for seed in range(self.N_MCMC)
            for model_name in list_model_names
        ]
        with ProcessPoolExecutor(
            max_workers=self.max_workers, mp_context=mp.get_context("fork")
        ) as p:
            list_results = list(
                tqdm(
                    p.map(_extract_mle_results_1_pix, list_params),
                    total=len(list_params),
                )
            )

        path_file = f"{self.path_data_csv_out_optim_mle}/results_MLE.csv"
        df_results_mle = pd.DataFrame(list_results)
        df_results_mle.to_csv(path_file)
        return

    def evaluate_mle(
        self,
        list_model_names: List[str],
        scaler: Scaler,
        estimator_plot: PlotsEstimator,
        Theta_true_scaled: Optional[np.ndarray] = None,
    ):
        for model_name in list_model_names:
            Theta_MLE, df_mle_final = self.read_estimator(model_name)
            Theta_MLE_scaled = self.scaler.from_lin_to_scaled(Theta_MLE)

            perf_saver = EstimatorPerfSaver()
            if Theta_true_scaled is not None:
                mse_whole = perf_saver.compute_MSE(Theta_MLE_scaled, Theta_true_scaled)
                snr_whole = perf_saver.compute_SNR(Theta_MLE_scaled, Theta_true_scaled)
            else:
                mse_whole = None
                snr_whole = None

            objective_whole = df_mle_final["objective"].sum()

            # save estimator
            perf_saver.save_estimator_performance(
                self.path_data_csv_out_mle,
                "MLE",
                model_name,
                mse_whole,
                snr_whole,
                objective_whole,
            )
            if self.N > 1:  # and self.index_arr.size > 1:
                folder_path_inter = f"{self.path_img}/estimators"
                folder_path = f"{folder_path_inter}/{model_name}"
                folder_path_MLE = f"{folder_path}/MLE"
                for path_ in [folder_path_inter, folder_path, folder_path_MLE]:
                    if not os.path.isdir(path_):
                        os.mkdir(path_)

                estimator_plot.plot_estimator()
                estimator_plot.plot_estimator(
                    Theta_MLE,
                    "MLE",
                    folder_path_MLE,
                    model_name,
                )

                # self.plot_map_nll_of_estimator(
                #     Theta_MLE_scaled,
                #     "MLE",
                #     model_name,
                # )

        return

    def plot_objectives(self, list_model_names: List[str]):
        global _plot_objectives

        def _plot_objectives(dict_input: dict) -> bool:
            n = dict_input["n"]
            model_name = dict_input["model_name"]

            folder_path_inter = f"{self.path_img}/objective"
            folder_path_inter2 = f"{folder_path_inter}/{model_name}"
            folder_path = f"{folder_path_inter2}/MLE_objectives"

            list_objective_all_seeds = np.zeros(
                (self.N_MCMC, self.effective_T_MLE),
            )

            # read objectives
            for seed in range(self.N_MCMC):
                path_results = f"{self.path_raw}/{model_name}/"
                path_results += f"opti_MLE_{seed}/pixel_{n}"

                with h5py.File(f"{path_results}/mc_chains.hdf5", "r") as f:
                    list_objective_seed = np.array(f["list_objective"])  # (T,)
                    assert list_objective_seed.shape == (self.effective_T_MLE,)

                list_objective_all_seeds[seed] += list_objective_seed

            # plot objectives
            plt.figure(figsize=(8, 6))
            plt.title(f"Objective evolution during optimization for pixel {n}")
            for seed in range(self.N_MCMC):
                plt.plot(
                    range(self.effective_T_MLE),
                    list_objective_all_seeds[seed, :],
                )
            if list_objective_all_seeds.max() <= 0:
                plt.yscale("linear")
            elif list_objective_all_seeds.min() <= 0:
                plt.yscale("symlog")
            else:
                plt.yscale("log")
            plt.grid()
            # plt.tight_layout()
            plt.savefig(
                f"{folder_path}/objective_pixel_{n}.PNG",
                bbox_inches="tight",
            )
            plt.close()

            return

        # * global function
        list_params = [
            {"n": n, "seed": seed, "model_name": model_name}
            for n in range(self.N)
            for seed in range(self.N_MCMC)
            for model_name in list_model_names
        ]
        with ProcessPoolExecutor(
            max_workers=self.max_workers, mp_context=mp.get_context("fork")
        ) as p:
            _ = list(
                tqdm(
                    p.map(_plot_objectives, list_params),
                    total=len(list_params),
                )
            )
        return

    def main(
        self,
        list_model_names: List[str],
        scaler: Scaler,
        estimator_plot: PlotsEstimator,
        Theta_true_scaled: Optional[np.ndarray] = None,
    ) -> None:
        """Note: list_model_names = list(self.dict_posteriors.keys())"""

        print("starting MLE results extraction.")
        self.extract_mle_results(list_model_names, scaler, Theta_true_scaled)
        self.evaluate_mle(list_model_names, scaler, estimator_plot)
        self.plot_objectives(list_model_names)
        print("MLE results extraction done")
        return



from typing import Dict, Optional, Tuple, Union

import numpy as np


class Posterior:

    __slots__ = (
        "D",
        "L",
        "N",
        "likelihood",
        "prior_spatial",
        "prior_indicator",
        "dict_sites",
    )

    def __init__(
        self,
        D: int,
        L: int,
        N: int,
        likelihood,
        prior_spatial=None,
        prior_indicator=None,
        separable: bool = True,
        dict_sites: Optional[Dict[int, np.ndarray]] = None,
    ):
        self.D = D
        """int: number of distinct physical parameters"""

        self.L = L
        """int: number of observables per pixel"""

        self.N = N
        """int: number of pixels"""

        self.likelihood = likelihood
        """Likelihood: data-fidelity term"""

        self.prior_spatial = prior_spatial
        """SpatialPrior: spatial prior term"""

        self.prior_indicator = prior_indicator
        """SmoothIndicatorPrior: prior term encoding validity intervals"""

        self.dict_sites = {}
        """dict[int, np.ndarray]: sites for pixels to be sampled in parallel in the MTM-chromoatic Gibbs kernel"""
        if dict_sites is not None:
            self.dict_sites = dict_sites
        elif self.prior_spatial is not None:
            self.dict_sites = self.prior_spatial.dict_sites
        elif separable is True:
            self.dict_sites = {0: np.arange(self.N)}
        else:
            self.dict_sites = {n: np.array([n]) for n in range(self.N)}

        return

    def partial_neglog_pdf_priors(
        self,
        Theta: np.ndarray,
        idx_pix: np.ndarray,
        list_pixel_candidates: np.ndarray,
        spatial_weights: Optional[np.ndarray] = None,
        use_indicator_prior: bool = True,
        use_spatial_prior: bool = True
        # compute_indicator: bool = False,
    ) -> np.ndarray:
        r"""computes the neg log-prior when only one pixel is modified

        Parameters
        ----------
        Theta : np.ndarray of shape (N, D)
            current iterate
        idx_pix : int
            the index of the pixel to consider (0 <= idx_pix <= N - 1)
        list_pixel_candidates : np.ndarray of shape (N_candidates, D)
            the list of all candidates for pixel idx_pi
        spatial_weights : Optional[np.ndarray], optional
            vector of shape (D,) containing the weights of the spatial prior, by default None
        use_indicator_prior : bool, optional
            wether to use the indicator prior term, by default True
        use_spatial_prior : bool, optional
            wether to use the spatial prior term, by default True

        Returns
        -------
        np.ndarray of shape (N_candidates,)
            the negative log-prior of the candidates
        """
        n_pix, k_mtm, D = list_pixel_candidates.shape
        assert n_pix == idx_pix.size
        assert D == Theta.shape[1]

        nl_priors = np.zeros((n_pix, k_mtm))
        if self.prior_spatial is not None and use_spatial_prior:
            nl_priors_spatial = self.prior_spatial.neglog_pdf_one_pix(
                Theta, idx_pix, list_pixel_candidates, spatial_weights
            )  # (n_pix, k_mtm)
            nl_priors += nl_priors_spatial

        if self.prior_indicator is not None and use_indicator_prior:
            list_pixel_candidates_reshaped = list_pixel_candidates.reshape(
                (n_pix * k_mtm, D)
            )
            nl_priors_indicator = self.prior_indicator.neglog_pdf_one_pix(
                list_pixel_candidates_reshaped
            )
            nl_priors += nl_priors_indicator.reshape((n_pix, k_mtm))
        return nl_priors

    def neglog_pdf_priors(
        self,
        Theta: np.ndarray,
        full: bool = False,
    ) -> Union[float, np.ndarray]:
        if full:
            nl_priors = np.zeros((self.N, self.L))
        else:
            nl_priors = 0.0

        if self.prior_spatial is not None:
            nl_prior_spatial = self.prior_spatial.neglog_pdf(Theta, pixelwise=full)
            if full:
                # nl_prior_spatial has shape (N, D), which needs to be
                # converted to (N, D)
                nl_prior_spatial = np.sum(nl_prior_spatial, axis=1)  # (N,)
                nl_priors += nl_prior_spatial[:, None]  # (N, D)
            else:
                nl_priors += np.sum(nl_prior_spatial)

        if self.prior_indicator is not None:
            nl_prior_ind = self.prior_indicator.neglog_pdf(Theta, pixelwise=full)
            if full:
                nl_priors += nl_prior_ind[:, None]
            else:
                nl_priors += np.sum(nl_prior_ind)

        return nl_priors

    def neglog_pdf(
        self,
        Theta: np.ndarray,
        forward_map_evals: dict,
        nll_utils: dict,
        full: bool = False,
    ) -> float:
        if full:
            out = np.zeros((self.N, self.L))
        else:
            out = 0.0

        out += self.likelihood.neglog_pdf(
            forward_map_evals,
            nll_utils,
            full=full,
        )

        out += self.neglog_pdf_priors(Theta, full=full)

        # assert np.sum(np.isnan(nll)) == 0, np.sum(np.isnan(nll))
        # assert np.sum(np.isnan(nl_priors)) == 0, np.sum(np.isnan(nl_priors))
        return out

    def grad_neglog_pdf(
        self,
        Theta: np.ndarray,
        forward_map_evals: dict[str, Union[float, np.ndarray]],
        nll_utils: dict[str, Union[float, np.ndarray]],
    ) -> np.ndarray:
        grad_ = self.likelihood.gradient_neglog_pdf(
            forward_map_evals, nll_utils
        )  # (N, D)
        # assert grad_.shape == (self.N, self.D), grad_nll.shape

        if self.prior_spatial is not None:
            # grad_nl_prior_spatial = self.prior_spatial.gradient_neglog_pdf(Theta)
            # assert grad_nl_prior_spatial.shape == (self.N, self.D)
            # assert (
            #     np.sum(np.isnan(grad_nl_prior_spatial)) == 0
            # ), f"nan grad prior spatial {np.sum(np.isnan(grad_nl_prior_spatial))}"
            grad_ += self.prior_spatial.gradient_neglog_pdf(Theta)

        if self.prior_indicator is not None:
            # grad_nl_prior_indicator = self.prior_indicator.gradient_neglog_pdf(Theta)
            # assert grad_nl_prior_indicator.shape == (self.N, self.D)
            # assert (
            #     np.sum(np.isnan(grad_nl_prior_indicator)) == 0
            # ), f"nan grad prior indicator {np.sum(np.isnan(grad_nl_prior_indicator))}"
            grad_ += self.prior_indicator.gradient_neglog_pdf(Theta)

        grad_ = np.nan_to_num(grad_)
        return grad_

    def hess_diag_neglog_pdf(
        self,
        Theta: np.ndarray,
        forward_map_evals: dict,
        nll_utils: dict,
    ) -> np.ndarray:
        hess_diag = self.likelihood.hess_diag_neglog_pdf(
            forward_map_evals, nll_utils
        )  # (N, D)
        # assert hess_diag.shape == (self.N, self.D)

        if self.prior_spatial is not None:
            # hess_diag_nl_prior_spatial = self.prior_spatial.hess_diag_neglog_pdf(Theta)
            # assert np.sum(np.isnan(hess_diag_nl_prior_spatial)) == 0
            # assert hess_diag_nl_prior_spatial.shape == (self.N, self.D)
            hess_diag += self.prior_spatial.hess_diag_neglog_pdf(Theta)

        if self.prior_indicator is not None:
            # hess_diag_nl_prior_indicator = self.prior_indicator.hess_diag_neglog_pdf(Theta)
            # assert np.sum(np.isnan(hess_diag_nl_prior_indicator)) == 0
            # assert hess_diag_nl_prior_indicator.shape == (self.N, self.D)
            hess_diag += self.prior_indicator.hess_diag_neglog_pdf(Theta)

        hess_diag = np.nan_to_num(hess_diag)
        return hess_diag

    def compute_all_for_saver(
        self,
        Theta: np.ndarray,
        forward_map_evals: dict[str, Union[float, np.ndarray]],
        nll_utils: dict,
    ) -> Tuple[dict[str, Union[float, np.ndarray]], np.ndarray]:
        """computes negative log pdf of likelihood, priors and posterior (detailed values to be saved, not to be used in sampling)

        Parameters
        ----------
        Theta : np.ndarray of shape (N, D)
            current iterate
        forward_map_evals : dict[str, Union[float, np.ndarray]]
            output of the ``likelihood.evaluate_all_forward_map()`` method
        nll_utils : [str, Union[float, np.ndarray]]
            output of the ``likelihood.evaluate_all_nll_utils()`` method

        Returns
        -------
        dict[str, Union[float, np.ndarray]]
            values to be saved
        """
        assert Theta.shape == (self.N, self.D)
        dict_objective = dict()

        nll_full = self.likelihood.neglog_pdf(
            forward_map_evals,
            nll_utils,
            full=True,
        )  # (N, L)

        assert isinstance(
            nll_full, np.ndarray
        ), "nll_full shoud be an array, check likelihood.neglog_pdf method"
        assert nll_full.shape == (
            self.N,
            self.L,
        ), f"nll_full with wrong shape. is {nll_full.shape}, should be {(self.N, self.L)}"

        dict_objective["nll"] = np.sum(nll_full)  # float

        if self.prior_spatial is not None:
            nl_prior_spatial = self.prior_spatial.neglog_pdf(Theta)
            dict_objective["nl_prior_spatial"] = nl_prior_spatial  # (D,)
        else:
            nl_prior_spatial = np.zeros((self.D,))

        if self.prior_indicator is not None:
            nl_prior_indicator = self.prior_indicator.neglog_pdf(Theta)
            dict_objective["nl_prior_indicator"] = nl_prior_indicator  # (D,)
        else:
            nl_prior_indicator = np.zeros((self.D,))

        nl_posterior = np.sum(nll_full) + np.sum(nl_prior_spatial)
        nl_posterior += np.sum(nl_prior_indicator)
        dict_objective["objective"] = nl_posterior

        return dict_objective, nll_full

    def compute_all(
        self,
        Theta: np.ndarray,
        forward_map_evals: dict = {},
        nll_utils: dict = {},
        compute_derivatives: bool = True,
        compute_derivatives_2nd_order: bool = True,
    ) -> dict:
        r"""compute negative log pdf and derivatives of the posterior distribution

        Parameters
        ----------
        Theta : np.ndarray of shape (N, D)
            current iterate
        forward_map_evals : dict[str, np.ndarray], optional
            output of the ``likelihood.evaluate_all_forward_map()`` method, by default {}
        nll_utils : dict[str, np.ndarray], optional
            output of the ``likelihood.evaluate_all_nll_utils()`` method, by default {}
        compute_derivatives : bool, optional
            wether to compte derivatives, by default True

        Returns
        -------
        dict[str, Union[float, np.ndarray]]
            negative log pdf and derivatives of the posterior distribution
        """
        assert np.sum(np.isnan(Theta)) == 0, np.sum(np.isnan(Theta))

        if forward_map_evals == {}:
            forward_map_evals = self.likelihood.evaluate_all_forward_map(
                Theta, compute_derivatives, compute_derivatives_2nd_order
            )

        if nll_utils == {}:
            nll_utils = self.likelihood.evaluate_all_nll_utils(
                forward_map_evals,
                None,
                compute_derivatives,
                compute_derivatives_2nd_order,
            )

        nll = self.likelihood.neglog_pdf(forward_map_evals, nll_utils)
        nll_utils["nll"] = nll  # float

        if self.prior_indicator is not None:
            nl_prior_indicator = self.prior_indicator.neglog_pdf(Theta)
        else:
            nl_prior_indicator = np.zeros((self.D,))
        nll_utils["nl_prior_indicator"] = nl_prior_indicator

        nlpdf_full = self.neglog_pdf(
            Theta,
            forward_map_evals,
            nll_utils,
            full=True,
        )  # (N, L)
        nlpdf_pix = np.sum(nlpdf_full, axis=1)

        iterate = {
            "Theta": Theta,
            "forward_map_evals": forward_map_evals,
            "nll_utils": nll_utils,
            "objective_pix": nlpdf_pix,
        }
        iterate["objective"] = np.sum(nlpdf_pix)
        if compute_derivatives:
            iterate["grad"] = self.grad_neglog_pdf(
                Theta,
                forward_map_evals,
                nll_utils,
            )
            if compute_derivatives_2nd_order:
                iterate["hess_diag"] = self.hess_diag_neglog_pdf(
                    Theta,
                    forward_map_evals,
                    nll_utils,
                )

        return iterate



import os
from typing import List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import colors
from scipy.stats import beta

from beetroots.inversion.plots.map_shaper import MapShaper
from beetroots.inversion.results.utils.abstract_util import ResultsUtil
from beetroots.sampler.abstract_sampler import Sampler


class ResultsBayesPvalues(ResultsUtil):
    r"""Bayesian model checking accounting for uncertainties on the p-value due to Monte Carlo evaluation.
    The method is described in :cite:t:`paludProblemesInversesTest2023a`.
    """

    __slots = (
        "model_name",
        "path_img",
        "path_data_csv_out",
        "N_MCMC",
        "N",
        "D",
    )

    CONFIDENCE_THRESHOLD_ALPHA = 0.05
    r"""..., denoted :math:`\alpha` in the article"""

    CONFIDENCE_THRESHOLD_DELTA = 0.1
    r"""..., denoted :math:`\delta` in the article"""

    ESS_OPTIM = Sampler.ESS_OPTIM * 1
    r"""number of random reproduced observations to draw to evaluate the model checking p-value for optimization procedures"""

    def __init__(
        self,
        model_name: str,
        chain_type: str,
        path_img: str,
        path_data_csv_out: str,
        N_MCMC: int,
        N: int,
        D_sampling: int,
        plot_ESS: bool,
    ):
        """

        Parameters
        ----------
        model_name : str
            _description_
        chain_type : str
            _description_
        path_img : str
            _description_
        path_data_csv_out : str
            _description_
        N_MCMC : int
            _description_
        N : int
            _description_
        D_sampling : int
            _description_
        plot_ESS : bool
            _description_
        """
        assert chain_type in ["mcmc", "optim_map", "optim_mle"]

        self.model_name = model_name
        self.chain_type = chain_type
        self.path_img = path_img
        self.path_data_csv_out = path_data_csv_out

        self.N_MCMC = N_MCMC
        self.N = N
        self.D_sampling = D_sampling

        self.plot_ESS = plot_ESS

    def read_data(self) -> Tuple[Optional[pd.DataFrame], pd.DataFrame]:
        # ess
        if self.chain_type == "mcmc":
            if self.plot_ESS:
                path_file = f"{self.path_data_csv_out}/"
                path_file += f"estimation_ESS_{self.model_name}.csv"

                df_ess_model = pd.read_csv(path_file, index_col=["n", "d"])
                df_ess_model = df_ess_model.sort_index()
                df_ess_model = df_ess_model.reset_index(drop=False)

            else:
                df_ess_model = None
        else:
            list_dicts = [
                {"n": n, "d": d, "ess": self.ESS_OPTIM, "seed": "overall"}
                for n in range(self.N)
                for d in range(self.D_sampling)
            ]
            df_ess_model = pd.DataFrame.from_records(list_dicts)
            df_ess_model = df_ess_model.set_index(["n", "d"])
            df_ess_model = df_ess_model.sort_index().reset_index(drop=False)

        # estimated pval
        path_file = f"{self.path_data_csv_out}/"
        path_file += f"p_value_llh_{self.model_name}.csv"

        df_pval = pd.read_csv(path_file, index_col=["seed", "n"])
        df_pval = df_pval.sort_index().reset_index(drop=False)

        assert df_ess_model is None or len(df_ess_model) == self.N * self.D_sampling
        assert len(df_pval) == self.N * self.N_MCMC
        return df_ess_model, df_pval

    def create_folders(self) -> str:
        folder_path_inter = f"{self.path_img}/p-values"
        folder_path = f"{folder_path_inter}/{self.model_name}"
        for path_ in [folder_path_inter, folder_path]:
            if not os.path.isdir(path_):
                os.mkdir(path_)

        return folder_path

    def main(
        self,
        list_idx_sampling: List[int],
        map_shaper: Optional[MapShaper],
    ) -> None:

        if self.N < 1:
            msg = "this function should only be called when N > 1 "
            msg += "to avoid 1-pixel maps"
            raise ValueError(msg)

        df_ess_model, df_pval = self.read_data()
        folder_path = self.create_folders()

        print("starting Bayesian p-value plots")

        if self.plot_ESS:
            assert df_ess_model is not None
            df_ess_overall = df_ess_model[(df_ess_model["seed"] == "overall")]
            df_ess = df_ess_overall.groupby("n")[["ess"]].min()
            df_ess = df_ess.sort_index()
            ess = df_ess["ess"].values

        list_decisions = []

        for seed in range(self.N_MCMC):
            df_pval_seed = df_pval[df_pval["seed"] == seed]
            df_pval_seed = df_pval_seed.sort_values("n")
            pval_estim = df_pval_seed["p-value-llh"].values

            decision_estim_seed = np.where(
                pval_estim < self.CONFIDENCE_THRESHOLD_ALPHA,
                -1,
                1,
            )

            if self.plot_ESS:
                proba_reject = np.zeros((self.N,))
                for n in range(self.N):
                    rv = beta(
                        1 + ess[n] * pval_estim[n],
                        1 + ess[n] * (1 - pval_estim[n]),
                    )
                    proba_reject[n] = rv.cdf(self.CONFIDENCE_THRESHOLD_ALPHA)

                decision_pval_bayes = np.where(
                    proba_reject > 1 - self.CONFIDENCE_THRESHOLD_DELTA,
                    -1,
                    np.where(
                        proba_reject < self.CONFIDENCE_THRESHOLD_DELTA,
                        1,
                        0,
                    ),
                )
            else:
                proba_reject = [None for n in range(self.N)]
                decision_pval_bayes = [None for n in range(self.N)]

            list_decisions += [
                {
                    "seed": seed,
                    "n": n,
                    "pval_estim": pval_estim[n],
                    "decision_estim_seed": decision_estim_seed[n],
                    "proba_reject": proba_reject[n],
                    "decision_pval_bayes": decision_pval_bayes[n],
                }
                for n in range(self.N)
            ]

            if map_shaper is not None:
                pval_estim_seed_shaped = map_shaper.from_vector_to_map(pval_estim)
                decision_estim_seed_shaped = map_shaper.from_vector_to_map(
                    decision_estim_seed
                )
                if proba_reject[0] is not None:
                    proba_reject_shaped = map_shaper.from_vector_to_map(proba_reject)
                    decision_pval_bayes_shaped = map_shaper.from_vector_to_map(
                        decision_pval_bayes
                    )

                plt.figure(figsize=(8, 6))
                title = "Estimated p-values"
                if self.N_MCMC > 1:
                    title += f" for seed={seed}"
                plt.title(title)

                plt.imshow(
                    pval_estim_seed_shaped,
                    norm=colors.LogNorm(),
                    cmap="jet",
                    origin="lower",
                )
                plt.colorbar()
                # plt.tight_layout()
                filename = f"{folder_path}/pval_estim_{self.chain_type}"
                filename += f"_seed{seed}.PNG"
                plt.savefig(filename, bbox_inches="tight")
                plt.close()

                #
                plt.figure(figsize=(8, 6))
                title = "Model rejection from estimated p-values"
                if self.N_MCMC > 1:
                    title += f" (seed={seed})"
                plt.title(title)

                plt.imshow(
                    decision_estim_seed_shaped,
                    cmap=colors.ListedColormap(["black", "white"]),
                    origin="lower",
                )

                colorbar = plt.colorbar(ticks=[-1 / 2, 1 / 2])
                colorbar.set_ticklabels(
                    ["reject", "reproduced"],
                    rotation=90,
                    va="center",
                )

                filename = f"{folder_path}/decision_pval_estim_{self.chain_type}"
                filename += f"_seed{seed}.PNG"
                plt.savefig(filename, bbox_inches="tight")
                # plt.tight_layout()
                plt.close()

                # * with bayesian computation of pval
                # Bayesian p-values
                if self.plot_ESS:
                    plt.figure(figsize=(8, 6))
                    plt.title(
                        r"Proba. reject per pixel $\mathbb{P}[p_n \leq \alpha]$",
                    )
                    plt.imshow(
                        proba_reject_shaped,
                        cmap="jet",
                        origin="lower",
                    )
                    plt.colorbar()
                    # plt.tight_layout()
                    plt.savefig(
                        f"{folder_path}/proba_reject_{self.chain_type}_seed{seed}.PNG",
                        bbox_inches="tight",
                    )
                    plt.close()

                    # decision process
                    plt.figure(figsize=(8, 6))
                    plt.title("Model rejection from p-values with uncertainties")
                    plt.imshow(
                        decision_pval_bayes_shaped,
                        cmap=colors.ListedColormap(["black", "grey", "white"]),
                        origin="lower",
                        norm=colors.Normalize(vmin=-1, vmax=1),
                    )

                    colorbar = plt.colorbar(ticks=[-2 / 3, 0, 2 / 3])
                    colorbar.set_ticklabels(
                        ["reject", "unclear", "no reject"],
                        rotation=90,
                        va="center",
                    )
                    plt.savefig(
                        f"{folder_path}/decision_from_bayes_pval_{self.chain_type}_seed{seed}.PNG",
                        bbox_inches="tight",
                    )
                    plt.close()

        df_pval = pd.DataFrame.from_records(list_decisions)
        df_pval.to_csv(f"{self.path_data_csv_out}/pvalues_analysis.csv")

        print("Bayesian p-value plots plots done")
        return



import abc


class Run(abc.ABC):
    r"""abstract class for inversion approach supervision, including sampling, optimization (MAP and MLE)"""

    @abc.abstractmethod
    def prepare_run(self):
        r"""prepares the inversion"""
        pass

    @abc.abstractmethod
    def run(self):
        r"""runs the inversion"""
        pass

    @abc.abstractmethod
    def main(self):
        r"""sequentially calls ``prepare_run`` and ``run``"""
        pass



from typing import Dict, List

import numpy as np
import pandas as pd


class MapShaper:
    """Defines the transformation from a vector to an map.
    Necessary for non-rectangular maps."""

    __slots__ = ("index_arr", "N")

    def __init__(self, df: pd.DataFrame) -> None:
        """

        Parameters
        ----------
        df : pandas.DataFrame
            should have (x, y) coordinates in the index and one "idx" column to define the transformation
        """
        index_arr = -np.ones(
            (
                df.index.get_level_values(0).max()
                - df.index.get_level_values(0).min()
                + 1,
                df.index.get_level_values(1).max()
                - df.index.get_level_values(1).min()
                + 1,
            )
        )

        min_theta = df.index.get_level_values(0).min()
        min_y = df.index.get_level_values(1).min()

        for (x, y) in df.index:
            index_arr[x - min_theta, y - min_y] = df.loc[(x, y), "idx"]

        # array of the indices of the vector in an image
        self.index_arr = index_arr.astype(int)
        r"""2D np.ndarray: `map` that contains the indices of the vector entries"""

        self.N = len(df)
        r"""int: number of pixels in map, i.e., dimension of vectors"""

    def from_vector_to_map(self, x: np.ndarray) -> np.ndarray:
        r"""applies a reshaping for a vector to plot it in 2D

        Parameters
        ----------
        x : numpy.array of shape (N,) or (N, -1)
            vector or tensor to be reshaped

        Returns
        -------
        x_shaped : numpy.array
            information of x shaped for plotting

        Raises
        ------
        ValueError
            if ``x`` has an invalid shape
        """
        if x.shape == (self.N,):
            x_shaped = np.empty_like(self.index_arr, dtype=np.float64)
            x_shaped[:, :] = np.nan
            for i in range(self.N):
                mask = self.index_arr == i
                x_shaped[mask] = x[i]

            return x_shaped.T

        if len(x.shape) == 2 and x.shape[0] == self.N:
            x_shaped = np.empty(
                (self.index_arr.shape[0], self.index_arr.shape[1], x.shape[1])
            )
            x_shaped[:, :] = np.nan
            for i in range(self.N):
                mask = self.index_arr == i
                x_shaped[mask] = x[i]
            return x_shaped.transpose((1, 0, 2))

        else:
            msg = f"invalid shape for x : {x.shape} "
            msg += f"(should be ({self.N},) or ({self.N}, -1))"
            raise ValueError(msg)

    def from_vector_idx_to_map_coords(
        self,
        idx: List[int],
    ) -> Dict[int, np.ndarray]:
        r"""transforms vector index in map coordinates

        Parameters
        ----------
        idx : numpy.array of shape (N,) or (N, -1)
            vector or tensor to be reshaped

        Returns
        -------
        x_shaped : numpy.array
            information of x shaped for plotting

        Raises
        ------
        ValueError
            if ``x`` has an invalid shape
        """
        x_idx = np.arange(self.N)
        x_idx_shaped = self.from_vector_to_map(x_idx)

        dict_coords = {k: None for k in idx}

        for i in range(x_idx_shaped.shape[0]):
            for j in range(x_idx_shaped.shape[1]):
                if not np.isnan(x_idx_shaped[i, j]):
                    pix_idx_val = int(x_idx_shaped[i, j])

                    for k, v in dict_coords.items():
                        if v is None and k == pix_idx_val:
                            dict_coords[k] = np.array([j, i])
                            break

        for k, v in dict_coords.items():
            assert v is not None, f"no coordinates found for pixel idx={k}"

        return dict_coords



from typing import Dict, List, Optional

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors

from beetroots.inversion.plots.map_shaper import MapShaper
from beetroots.inversion.plots.plots_2d import AbstractPlots2D


class PlotsEstimator(AbstractPlots2D):
    r"""utilitary class that draws and saves maps related to the inference results."""

    __slots__ = (
        "map_shaper",
        "list_names",
        "lower_bounds_lin",
        "upper_bounds_lin",
        "list_idx_sampling",
        "pixels_of_interest_names",
        "pixels_of_interest_coords",
    )

    def __init__(
        self,
        map_shaper: MapShaper,
        list_names: List[str],
        lower_bounds_lin: np.ndarray,
        upper_bounds_lin: np.ndarray,
        list_idx_sampling: List[int],
        pixels_of_interest: Dict[int, str] = {},
    ):
        self.map_shaper = map_shaper
        r"""MapShaper: defines the transformation from vectors to 2D maps"""

        self.list_names = list_names
        r"""list: physical parameter names"""

        self.lower_bounds_lin = lower_bounds_lin
        r"""1D np.ndarray: contains the lower bounds in linear scale on the physical parameters"""
        self.upper_bounds_lin = upper_bounds_lin
        r"""1D np.ndarray: contains the upper bounds in linear scale on the physical parameters"""

        self.list_idx_sampling = list_idx_sampling
        r"""1D np.ndarray: contains the indices of the physical parameters to be sampled"""

        self.pixels_of_interest_names = pixels_of_interest
        r"""dict: (coordinate, name) pair of some user-informed pixels to be highlighted. These pixels will be outlines with a black square in figures."""

        coords = map_shaper.from_vector_idx_to_map_coords(
            list(pixels_of_interest.keys())
        )
        self.pixels_of_interest_coords = coords
        r"""list: coordinates of some user-informed pixels to be highlighted. These pixels will be outlines with a black square in figures."""

    def plot_estimator(
        self,
        Theta_estimated: np.ndarray,
        estimator_name: str,
        folder_path: str,
        model_name: Optional[str] = "",
    ):
        """plots and saves the 2D map of an estimated physical parameter

        Parameters
        ----------
        Theta_estimated : np.ndarray of shape (N, D)
            vector of the estimated physical parameter
        estimator_name : str
            name of the estimator, e.g., "MLE", "MAP" or "MMSE"
        folder_path : str
            path to the folder where the figure is to be saved
        model_name : Optional[str], optional
            name of the model (not used here, kept for compatibility), by default ""
        """
        Theta_estimated_plot = self.map_shaper.from_vector_to_map(Theta_estimated)
        for d, name in enumerate(self.list_names):
            vmin = self.lower_bounds_lin[d] / 1.1
            vmax = self.upper_bounds_lin[d] * 1.1

            x_estimator_d_plot = Theta_estimated_plot[:, :, d] * 1

            title = f"{estimator_name} for {name}"
            if d not in self.list_idx_sampling:
                title += " (fixed, not estimated)"

            if vmin > 0:
                plt.figure(figsize=(8, 6))
                plt.title(title)
                plt.imshow(
                    x_estimator_d_plot,
                    norm=colors.LogNorm(vmin, vmax),
                    origin="lower",
                    cmap="jet",
                )
                plt.colorbar()
                self._draw_rect_on_pixels_of_interest()
                # plt.tight_layout()

                filename = f"{folder_path}/{estimator_name}_{d}"
                filename = filename.replace("%", "").replace(".", "p")
                filename = filename.replace(" ", "_")
                filename += ".PNG"
                plt.savefig(
                    filename,
                    bbox_inches="tight",
                )
                plt.close()

            # * same in linear scale
            plt.figure(figsize=(8, 6))
            plt.title(title)
            plt.imshow(
                x_estimator_d_plot,
                origin="lower",
                cmap="jet",
            )
            plt.colorbar()
            self._draw_rect_on_pixels_of_interest()

            filename = f"{folder_path}/"
            filename += f"{estimator_name}_linscale_{d}"
            filename = filename.replace("%", "").replace(".", "p")
            filename = filename.replace(" ", "_")
            filename += ".PNG"

            plt.savefig(filename, bbox_inches="tight")
            plt.close()

    def plot_estimator_u(
        self,
        u_estimated: np.ndarray,
        estimator_name: str,
        folder_path: str,
        model_name: Optional[str] = "",
        list_lines: List[str] = [],
    ):
        """Only used in hierarchical models. The sampling of such model is not implemented."""
        u_estimated_plot = self.map_shaper.from_vector_to_map(u_estimated)

        for ell, line in enumerate(list_lines):
            u_estimator_ell_plot = u_estimated_plot[:, :, ell] * 1

            plt.figure(figsize=(8, 6))
            plt.title(f"{estimator_name} for {line}")
            plt.imshow(
                u_estimator_ell_plot,
                norm=colors.LogNorm(),
                origin="lower",
                cmap="jet",
            )
            plt.colorbar()
            self._draw_rect_on_pixels_of_interest()
            # plt.tight_layout()

            filename = f"{folder_path}/{estimator_name}_{ell}_{line}"
            filename = filename.replace("%", "").replace(".", "p")
            filename += ".PNG"
            plt.savefig(filename, bbox_inches="tight")
            plt.close()

            # * same in linear scale
            plt.figure(figsize=(8, 6))
            plt.title(f"{estimator_name} for {line}")
            plt.imshow(u_estimator_ell_plot, origin="lower", cmap="jet")
            plt.colorbar()

            filename = f"{folder_path}/"
            filename += f"{estimator_name}_linscale_{ell}_{line}"
            filename = filename.replace("%", "").replace(".", "p")
            filename += ".PNG"

            plt.savefig(filename, bbox_inches="tight")
            plt.close()

    def plot_CI_size(
        self,
        Theta_ci_size: np.ndarray,
        CI_name: str,
        folder_path: str,
    ) -> None:
        r"""plots the map of credibility interval sizes for a physical parameter

        Parameters
        ----------
        Theta_ci_size : np.ndarray of shape (N, D)
            vector of credibility interval sizes for the D physical parameters
        CI_name : str
            name of the credibility interval, e.g., "95\%" or "99\%"
        folder_path : str
            path to the folder where the figure is to be saved
        """
        Theta_ci_size_plot = self.map_shaper.from_vector_to_map(Theta_ci_size)
        for d, name in enumerate(self.list_names):
            if d in self.list_idx_sampling:
                x_ci_size_d_plot = Theta_ci_size_plot[:, :, d] * 1

                if self.upper_bounds_lin[-1] - self.lower_bounds_lin[-1] > 20:
                    vmin = 1
                    vmax = None
                    # vmax = self.upper_bounds_lin[d] / self.lower_bounds_lin[d]
                else:
                    vmin = 0
                    vmax = None
                    # vmax = self.upper_bounds_lin[d] - self.lower_bounds_lin[d]

                plt.figure(figsize=(8, 6))
                plt.title(f"{CI_name} for {name}")
                plt.imshow(
                    x_ci_size_d_plot,
                    norm=colors.LogNorm(vmin, vmax),
                    origin="lower",
                    cmap="jet",
                )
                plt.colorbar()
                self._draw_rect_on_pixels_of_interest()

                filename = f"{folder_path}/{CI_name}_d{d}.PNG"
                filename = filename.replace("%", "").replace(" ", "_")
                plt.savefig(filename, bbox_inches="tight")
                plt.close()

                # * same in linear scale
                plt.figure(figsize=(8, 6))
                plt.title(f"{CI_name} for {name}")
                plt.imshow(
                    x_ci_size_d_plot,
                    origin="lower",
                    cmap="jet",
                )
                plt.colorbar()
                self._draw_rect_on_pixels_of_interest()

                filename = f"{folder_path}/{CI_name}_linscale_d{d}.PNG"
                filename = filename.replace("%", "").replace(" ", "_")
                plt.savefig(filename, bbox_inches="tight")
                plt.close()

        return

    def plot_CI_size_u(
        self,
        u_ci_size,
        CI_name,
        folder_path: str,
        model_name: Optional[str] = "",
        list_lines: List[str] = [],
    ):
        """

        Only used in hierarchical models. The sampling of such model is not implemented.

        """

        u_ci_size_plot = self.map_shaper.from_vector_to_map(u_ci_size)
        for ell, line in enumerate(list_lines):
            u_ci_size_d_plot = u_ci_size_plot[:, :, ell] * 1

            plt.figure(figsize=(8, 6))
            plt.title(f"{CI_name} for {line}")
            plt.imshow(
                u_ci_size_d_plot,
                norm=colors.LogNorm(),
                origin="lower",
                cmap="jet",
            )
            plt.colorbar()
            self._draw_rect_on_pixels_of_interest()

            filename = f"{folder_path}/{CI_name}_{ell}.PNG"
            filename = filename.replace("%", "")
            plt.savefig(filename, bbox_inches="tight")
            plt.close()

            # * same in linear scale
            plt.figure(figsize=(8, 6))
            plt.title(f"{CI_name} for {line}")
            plt.imshow(
                u_ci_size_d_plot,
                origin="lower",
                cmap="jet",
            )
            plt.colorbar()
            self._draw_rect_on_pixels_of_interest()

            filename = f"{folder_path}/{CI_name}_linscale_{ell}.PNG"
            filename = filename.replace("%", "")
            plt.savefig(filename, bbox_inches="tight")
            plt.close()



import os
import sys

import numpy as np
import yaml

from beetroots.approx_optim.nn_bo import ApproxParamsOptimNNBO

if __name__ == "__main__":

    if len(sys.argv) < 2:
        print("Please provide the name of the YAML file as an argument.")
        sys.exit(1)

    path_yaml_file = sys.argv[1]
    path_data_cloud, yaml_file = os.path.split(path_yaml_file)

    # step 1: read ``.yaml`` input file
    with open(os.path.abspath(f"{path_data_cloud}/{yaml_file}")) as f:
        params = yaml.safe_load(f)

    # step 2: run the Bayesian optimization
    approx_optim = ApproxParamsOptimNNBO(
        params["list_lines"],
        sigma_a=params["sigma_a"],
        sigma_m=np.log(params["sigma_m_float_linscale"]),
        **params["simu_init"],
    )
    approx_optim.main(**params["main_params"])



import multiprocessing as mp
import os
import warnings
from concurrent.futures import ProcessPoolExecutor
from typing import Dict, List

import h5py
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from beetroots.inversion.plots import readable_line_names
from beetroots.inversion.results.utils.abstract_util import ResultsUtil
from beetroots.modelling.forward_maps.abstract_base import ForwardMap
from beetroots.space_transform.abstract_transform import Scaler


class ResultsDistributionComparisonYandFTheta(ResultsUtil):

    __slots__ = (
        "model_name",
        "path_img",
        "path_data_csv_out_mcmc",
        "N_MCMC",
        "T_MC",
        "T_BI",
        "freq_save",
        "effective_len_mc",
        "N",
        "D",
    )

    def __init__(
        self,
        model_name: str,
        path_img: str,
        path_data_csv_out_mcmc: str,
        N_MCMC: int,
        T_MC: int,
        T_BI: int,
        freq_save: int,
        N: int,
        D: int,
        list_idx_sampling: List[int],
        max_workers: int,
    ):
        self.model_name = model_name
        self.path_img = path_img
        self.path_data_csv_out_mcmc = path_data_csv_out_mcmc

        self.N_MCMC = N_MCMC
        self.T_MC = T_MC
        self.T_BI = T_BI
        self.freq_save = freq_save
        self.effective_len_mc = (T_MC - T_BI) // freq_save

        self.N = N
        self.D = D

        self.list_idx_sampling = list_idx_sampling
        self.D_sampling = len(list_idx_sampling)

        self.max_workers = max_workers

    def read_data(self) -> np.ndarray:
        path_file = f"{self.path_data_csv_out_mcmc}/"
        path_file += f"estimation_Theta_{self.model_name}_MAP_mcmc.csv"

        df_Theta_MAP_lin = pd.read_csv(path_file, index_col=["n", "d"])
        df_Theta_MAP_lin = df_Theta_MAP_lin.sort_index()
        Theta_MAP_lin_full = df_Theta_MAP_lin["Theta_MAP_mcmc"].values.reshape(
            (self.N, self.D)
        )

        return Theta_MAP_lin_full

    def create_folders(self) -> str:
        folder_path_inter = f"{self.path_img}/distri_comp_yspace"
        folder_path = f"{folder_path_inter}/{self.model_name}"
        for path_ in [folder_path_inter, folder_path]:
            if not os.path.isdir(path_):
                os.mkdir(path_)

        return folder_path

    def main(
        self,
        list_mcmc_folders: List[str],
        scaler: Scaler,
        forward_map: ForwardMap,
        y: np.ndarray,
        omega: np.ndarray,
        sigma_a: np.ndarray,
        sigma_m: np.ndarray,
        list_lines: List[str],
        name_list_lines: str,
        point_challenger: Dict = {},
    ) -> None:
        global _plot_one_distribution_comparison

        assert name_list_lines in ["fit", "valid"]

        N_samples_per_chain = self.effective_len_mc // 20
        N_samples = N_samples_per_chain * self.N_MCMC
        rng = np.random.default_rng(42)

        L = y.shape[1]

        folder_path = self.create_folders()

        forward_map.restrict_to_output_subset(list_lines)

        print("starting plot comparison of distributions of y and f(theta)")

        Theta_MAP_lin_full = self.read_data()
        Theta_MAP_scaled_full = scaler.from_lin_to_scaled(Theta_MAP_lin_full)

        Theta_MAP_scaled = Theta_MAP_scaled_full[:, self.list_idx_sampling]

        f_Theta_MAP = forward_map.evaluate(Theta_MAP_scaled)

        if len(point_challenger) > 0:
            x_challenger_full = scaler.from_lin_to_scaled(point_challenger["value"])
            x_challenger = x_challenger_full[:, self.list_idx_sampling]
            f_Theta_challenger = forward_map.evaluate(x_challenger)

        # try:
        #     Theta_MLE_lin, _ = self.read_MLE_from_csv_file(model_name)
        #     Theta_MLE_scaled = self.scaler.from_lin_to_scaled(Theta_MLE_lin)
        #     f_Theta_MLE = forward_map.evaluate(Theta_MLE_scaled)
        # except:
        #     f_Theta_MLE = None

        def add_label(violin, label, list_labels):
            color = violin["bodies"][0].get_facecolor().flatten()
            list_labels.append((mpatches.Patch(color=color), label))
            return list_labels

        def _plot_one_distribution_comparison(n):
            list_theta_n_lin = np.zeros(
                (self.N_MCMC, N_samples_per_chain, self.D_sampling)
            )
            for seed, mc_path in enumerate(list_mcmc_folders):
                list_t = list(
                    rng.choice(
                        a=np.arange(
                            self.T_BI // self.freq_save,
                            self.T_MC // self.freq_save,
                        ),
                        size=N_samples_per_chain,
                        replace=False,
                    )
                )  # list
                list_t.sort()  # list

                with h5py.File(mc_path, "r") as f:
                    list_theta_n_lin[seed] = np.array(f["list_Theta"][list_t, n, :])

            list_theta_n_lin = list_theta_n_lin.transpose((2, 0, 1))
            list_theta_n_lin = list_theta_n_lin.reshape((self.D_sampling, N_samples)).T

            list_theta_n_lin_full = np.ones((N_samples, self.D))
            for idx_d, d in enumerate(self.list_idx_sampling):
                list_theta_n_lin_full[:, d] = list_theta_n_lin[:, idx_d]

            list_theta_n_scaled_full = scaler.from_lin_to_scaled(list_theta_n_lin_full)

            list_theta_n_scaled = list_theta_n_scaled_full[:, self.list_idx_sampling]

            list_f_Theta_n_lin = forward_map.evaluate(list_theta_n_scaled)  # (N_s, L)

            list_labels = []

            plt.figure(figsize=(12, 8))
            if self.N > 1:
                title = r"comparison of $f(\theta)$ and $y$ distributions for pixel"
                title += f" {n}"
                plt.title(title)
            else:
                plt.title(r"comparison of $f(\theta)$ and $y$ distributions")

            plt.xlabel("lines")
            plt.ylabel(r"$\log y$")

            n_std = 1

            l_y_add = None
            for ell in range(L):
                if y[n, ell] > omega[n, ell]:
                    list_theta = np.array([ell, ell, ell]) + 1
                    list_y = [
                        y[n, ell] - n_std * sigma_a[n, ell],
                        y[n, ell],
                        y[n, ell] + n_std * sigma_a[n, ell],
                    ]

                    (l_y_add,) = plt.plot(list_theta, list_y, "C2_-")

            l_y_multi = None
            for ell in range(L):
                if y[n, ell] > omega[n, ell]:
                    list_theta = np.array([ell, ell, ell]) + 1.15
                    list_y = [
                        np.exp(np.log(y[n, ell]) - (n_std * sigma_m[n, ell])),
                        y[n, ell],
                        np.exp(np.log(y[n, ell]) + (n_std * sigma_m[n, ell])),
                    ]

                    (l_y_multi,) = plt.plot(list_theta, list_y, "C1_-")

            list_labels = []
            if l_y_add is not None:
                list_labels.append((l_y_add, r"$Y \pm \sigma_a$"))
            if l_y_multi is not None:
                list_labels.append((l_y_multi, r"$Y \pm \sigma_m Y$"))

            list_labels = add_label(
                plt.violinplot(
                    list_f_Theta_n_lin,
                    positions=np.arange(1, L + 1) + 0.45,
                    widths=0.45,
                    showmeans=True,
                    showextrema=True,
                ),
                label=r"$f(Theta)$",
                list_labels=list_labels,
            )

            l_map = plt.scatter(
                np.arange(1, L + 1) + 0.45,
                f_Theta_MAP[n, :],
                marker="*",
                c="r",
                s=50,
            )
            list_labels += [(l_map, r"$f(\hat{x}_{MAP})$")]

            if len(point_challenger) > 0:
                l_challenger = plt.scatter(
                    np.arange(1, L + 1) + 0.45,
                    f_Theta_challenger[n, :],
                    marker="+",
                    c="k",
                    s=50,
                )
                list_labels += [(l_challenger, point_challenger["name"])]

            # if f_Theta_MLE is not None:
            #     l_mle = plt.scatter(
            #         np.arange(1, L + 1) + 0.45,
            #         f_Theta_MLE[n, :],
            #         marker="*",
            #         c="g",
            #         s=50,
            #     )
            #     list_labels += [(l_mle, r"$f(\hat{x}_{MLE})$")]

            l_omega = None
            for ell in range(L):
                if y[n, ell] <= omega[n, ell]:
                    (l_omega,) = plt.plot(
                        [1 + ell, 1 + ell + 0.8],
                        [omega[n, ell], omega[n, ell]],
                        "k--",
                    )

            if l_omega is not None:
                list_labels.append((l_omega, r"censor limit $\omega$"))

            list_lines_readable = [
                readable_line_names.lines_to_latex(line) for line in list_lines
            ]
            plt.xticks(np.arange(1, L + 1), list_lines_readable, rotation=90)
            plt.yscale("log")
            plt.ylim([min(0.1, list_f_Theta_n_lin.min() / 2), None])
            plt.grid()
            plt.legend(*zip(*list_labels))
            # plt.tight_layout()

            filename = f"{folder_path}/distribution_comparison_"
            filename += f"pix_{n}_{name_list_lines}.PNG"

            plt.savefig(filename, bbox_inches="tight")
            plt.close()
            return

        list_params = range(self.N)

        # set forward map device to cpu to make parallel computations possible
        device = forward_map.network.device * 1
        if "cuda" in forward_map.network.device:
            forward_map.network.set_device("cpu")

        # ? The parallel execution may fail on mac, even with the mp_context
        # ? argument. As I can't correct the error, in case of fail, I perform
        # ? the extraction in series, which is much slower.
        try:
            with ProcessPoolExecutor(
                max_workers=self.max_workers, mp_context=mp.get_context("fork")
            ) as p:
                _ = list(
                    tqdm(
                        p.map(_plot_one_distribution_comparison, list_params),
                        total=self.N,
                    )
                )

            # set back forward map device on its original value
            if "cuda" in device:
                forward_map.network.set_device(device)

        except:
            warnings.warn(
                "The parallel pixel-wise result extraction failed. Extracting in series instead."
            )

            # set back forward map device on its original value
            if "cuda" in device:
                forward_map.network.set_device(device)

            # if "cuda" in forward_map.network.device:
            for params in tqdm(list_params):
                _plot_one_distribution_comparison(params)

        print("plot comparison of distributions of y and f(theta) done")
        return



import os
import sys
from typing import List

import numpy as np
import yaml

from beetroots.approx_optim.nn_bo import ApproxParamsOptimNNBO
from beetroots.simulations.astro.observation.abstract_real_data import (
    SimulationRealData,
)


class ReadDataRealData(SimulationRealData):
    r"""implements an ``__init__`` method for :class:`.SimulationRealData`"""

    def __init__(self, list_lines: List[str]) -> None:
        """

        Parameters
        ----------
        list_lines : List[str]
            observables for which the likelihood approximation parameters are to be adjusted
        """
        self.list_lines_fit = list_lines


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("Please provide the name of the YAML file as an argument.")
        sys.exit(1)

    path_yaml_file = sys.argv[1]
    path_data_cloud, yaml_file = os.path.split(path_yaml_file)

    # step 1: read ``.yaml`` input file
    with open(os.path.abspath(f"{path_data_cloud}/{yaml_file}")) as f:
        params = yaml.safe_load(f)

    # step 2: read the additive noise standard deviations for the specified
    # lines
    data_reader = ReadDataRealData(params["list_lines"])
    (_, _, sigma_a, _, _, _, _) = data_reader.setup_observation(
        data_int_path=f"{path_data_cloud}/{params['filename_int']}",
        data_err_path=f"{path_data_cloud}/{params['filename_err']}",
        save_obs=False,
    )

    # step 3: run the Bayesian optimization
    approx_optim = ApproxParamsOptimNNBO(
        list_lines=params["list_lines"],
        sigma_a=sigma_a,
        sigma_m=np.log(params["sigma_m_float_linscale"]),
        **params["simu_init"],
    )
    approx_optim.main(**params["main_params"])



import os
from typing import List

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import colors

from beetroots.inversion.plots.map_shaper import MapShaper
from beetroots.inversion.results.utils.abstract_util import ResultsUtil


class ResultsESS(ResultsUtil):

    __slots = (
        "model_name",
        "path_img",
        "path_data_csv_out_mcmc",
        "N",
        "D",
    )

    def __init__(
        self,
        model_name: str,
        path_img: str,
        path_data_csv_out_mcmc: str,
        N: int,
        D_sampling: int,
    ):
        self.model_name = model_name
        self.path_img = path_img
        self.path_data_csv_out_mcmc = path_data_csv_out_mcmc

        self.N = N
        self.D_sampling = D_sampling

    def read_data(self) -> pd.DataFrame:
        path_file = f"{self.path_data_csv_out_mcmc}/"
        path_file += f"estimation_ESS_{self.model_name}.csv"

        df_ess_model = pd.read_csv(path_file, index_col=["n", "d"])
        df_ess_model = df_ess_model.sort_index().reset_index(drop=False)

        assert len(df_ess_model) == self.N * self.D_sampling
        return df_ess_model

    def create_folders(self) -> str:
        folder_path_inter = f"{self.path_img}/ess"
        folder_path = f"{folder_path_inter}/{self.model_name}"
        for path_ in [folder_path_inter, folder_path]:
            if not os.path.isdir(path_):
                os.mkdir(path_)

        return folder_path

    def main(
        self,
        map_shaper: MapShaper,
        list_names: List[str],
        list_idx_sampling: List[int],
    ) -> None:

        if self.N < 1:
            msg = "this function should only be called when N > 1 "
            msg += "to avoid 1-pixel maps"
            raise ValueError(msg)

        df_ess_model = self.read_data()
        folder_path = self.create_folders()

        print("starting ESS plots")

        for d in list_idx_sampling:
            df_ess_overall = df_ess_model[
                (df_ess_model["seed"] == "overall") & (df_ess_model["d"] == d)
            ]
            df_ess_overall = df_ess_overall.sort_values("n")

            ess_arr = df_ess_overall.loc[:, "ess"].values
            ess_arr_shaped = map_shaper.from_vector_to_map(ess_arr)

            plt.figure(figsize=(8, 6))
            plt.title(f"ESS per pixel for {list_names[d]}")
            plt.imshow(
                ess_arr_shaped,
                norm=colors.LogNorm(vmin=1.0),
                cmap="jet",
                origin="lower",
            )
            plt.colorbar()
            # plt.tight_layout()
            plt.savefig(f"{folder_path}/ESS_d{d}.PNG", bbox_inches="tight")
            plt.close()

        print("ESS plots done")
        return



import os
from typing import Dict, List, Optional, Union

import numpy as np

from beetroots.modelling.priors.spatial_prior_params import SpatialPriorParams
from beetroots.sampler.utils.my_sampler_params import MySamplerParams
from beetroots.simulations.abstract_simulation import Simulation
from beetroots.simulations.astro import data_validation
from beetroots.simulations.astro.forward_map.abstract_nn import SimulationNN
from beetroots.simulations.astro.observation.abstract_toy_case import SimulationToyCase
from beetroots.simulations.astro.posterior_type.abstract_direct import (
    SimulationMySampler,
)


class SimulationToyCaseNN(
    Simulation, SimulationToyCase, SimulationNN, SimulationMySampler
):
    __slots__ = (
        "path_output_sim",
        "path_img",
        "path_raw",
        "path_data_csv",
        "path_data_csv_in",
        "path_data_csv_out",
        "path_data_csv_out_mcmc",
        "path_data_csv_out_optim_map",
        "path_data_csv_out_optim_mle",
        "N",
        "D",
        "D_no_kappa",
        "L",
        "list_names",
        "list_names_plot",
        "cloud_name",
        "max_workers",
        "list_lines_fit",
        "list_lines_valid",
        "Theta_true_scaled",
        "map_shaper",
        "plots_estimator",
    )

    def setup(
        self,
        forward_model_name: str,
        force_use_cpu: bool,
        fixed_params: Dict[str, Optional[float]],
        is_log_scale_params: Dict[str, bool],
        #
        sigma_a_float: float,
        sigma_m_float: float,
        omega_float: float,
        #
        indicator_margin_scale: float,
        lower_bounds_lin: np.ndarray,
        upper_bounds_lin: np.ndarray,
        #
        with_spatial_prior: bool = True,
        spatial_prior_params: Union[None, SpatialPriorParams] = None,
        list_gaussian_approx_params: List[str] = [],
        list_mixing_model_params: List[str] = [],
    ):
        self.N = int(self.cloud_name.split("N")[1]) ** 2

        self.list_lines_valid = []

        scaler, forward_map = self.setup_forward_map(
            forward_model_name=forward_model_name,
            force_use_cpu=force_use_cpu,
            dict_fixed_params=fixed_params,
            dict_is_log_scale_params=is_log_scale_params,
        )

        sigma_a = sigma_a_float * np.ones((self.N, self.L))
        sigma_m = sigma_m_float * np.ones((self.N, self.L))
        omega = omega_float * np.ones((self.N, self.L))

        syn_map, y = self.setup_observation(
            scaler=scaler,
            forward_map=forward_map,
            sigma_a=sigma_a,
            sigma_m=sigma_m,
            omega=omega,
        )

        # run setup
        dict_posteriors, scaler, prior_indicator_1pix = self.setup_posteriors(
            scaler=scaler,
            forward_map=forward_map,
            y=y,
            sigma_a=sigma_a,
            sigma_m=sigma_m,
            omega=omega,
            syn_map=syn_map,
            with_spatial_prior=with_spatial_prior,
            spatial_prior_params=spatial_prior_params,
            indicator_margin_scale=indicator_margin_scale,
            lower_bounds_lin=lower_bounds_lin,
            upper_bounds_lin=upper_bounds_lin,
            list_gaussian_approx_params=list_gaussian_approx_params,
            list_mixing_model_params=list_mixing_model_params,
        )

        y_valid = None
        sigma_a_valid = None
        omega_valid = None

        return (
            dict_posteriors,
            scaler,
            prior_indicator_1pix,
            y_valid,
            sigma_a_valid,
            omega_valid,
        )

    def main(self, params: dict, path_data_cloud: str) -> None:
        if params["with_spatial_prior"]:
            spatial_prior_params = SpatialPriorParams(
                **params["spatial_prior"],
            )
        else:
            spatial_prior_params = None

        (
            dict_posteriors,
            scaler,
            prior_indicator_1pix,
            y_valid,  # None
            sigma_a_valid,  # None
            omega_valid,  # None
        ) = simulation.setup(
            **params["forward_model"],
            #
            sigma_a_float=params["sigma_a_float"],
            sigma_m_float=np.log(params["sigma_m_float_linscale"]),
            omega_float=3 * params["sigma_a_float"],
            #
            **params["prior_indicator"],
            #
            with_spatial_prior=params["with_spatial_prior"],
            spatial_prior_params=spatial_prior_params,
            #
            list_gaussian_approx_params=params["list_gaussian_approx_params"],
            list_mixing_model_params=[
                {"path_transition_params": f"{path_data_cloud}/{filename}"}
                for filename in params["mixing_model_params_filename"]
            ],
        )
        simulation.save_and_plot_setup(
            dict_posteriors,
            params["prior_indicator"]["lower_bounds_lin"],
            params["prior_indicator"]["upper_bounds_lin"],
            scaler,
        )
        # * Optim MAP
        if params["to_run_optim_map"]:
            simulation.inversion_optim_map(
                dict_posteriors=dict_posteriors,
                scaler=scaler,
                my_sampler_params=MySamplerParams(**params["sampling_params"]["map"]),
                can_run_in_parallel=params["forward_model"]["force_use_cpu"],
                **params["run_params"]["map"],
            )

        # * MCMC
        if params["to_run_mcmc"]:
            simulation.inversion_mcmc(
                dict_posteriors=dict_posteriors,
                scaler=scaler,
                my_sampler_params=MySamplerParams(**params["sampling_params"]["mcmc"]),
                can_run_in_parallel=params["forward_model"]["force_use_cpu"],
                **params["run_params"]["mcmc"],
            )
        return


if __name__ == "__main__":
    path_data_cloud = f"{os.path.dirname(os.path.abspath(__file__))}"
    path_data_cloud += "/../../../../data/toycases"

    params = SimulationToyCaseNN.load_params(path_data_cloud)

    simulation = SimulationToyCaseNN(**params["simu_init"], params=params)
    SimulationToyCaseNN.check_input_params_file(
        params,
        data_validation.schema,
    )

    simulation.main(
        params=params,
        path_data_cloud=path_data_cloud,
    )



"""Contains a multi-line polynomial regressor made to approximate the true forward model given a grid of simulations
"""
import os
import pickle
from typing import List, Optional, Sequence, Tuple, Union

import numba as nb
import numpy as np
import pandas as pd

from beetroots.modelling.forward_maps.abstract_exp import ExpForwardMap


def compute_grad_poly_params(
    pow_arr: np.ndarray, coeff_arr: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """computes the power array and the coefficient array of the gradient of the multivariate polynomial defined with the given power and coefficient arrays

    Parameters
    ----------
    pow_arr : np.ndarray of shape (n_coefs, D_no_kappa)
        array of powers of the initial multivariate polynomial
    coeff_arr : np.ndarray of shape (L, n_coefs)
        array of coeffients of the initial multivariate polynomial

    Returns
    -------
    pow_grad : np.ndarray of shape (D_no_kappa, n_coefs, D_no_kappa)
        array of powers of the gradient of multivariate polynomial
    coef_grad : np.ndarray of shape (D_no_kappa, L, n_coefs)
        array of coeffients of the gradient of multivariate polynomial
    """
    L, n_coefs = coeff_arr.shape
    n_coefs_2, D_no_kappa = pow_arr.shape
    assert n_coefs == n_coefs_2

    pow_grad = pow_arr[None, :, :] * np.ones((D_no_kappa, n_coefs, D_no_kappa))
    coef_grad = coeff_arr[None, :, :] * np.ones((D_no_kappa, L, n_coefs))

    for d_deriv in range(D_no_kappa):
        for i_coef in range(n_coefs):
            if pow_grad[d_deriv, i_coef, d_deriv] > 0:
                pow_grad[d_deriv, i_coef, d_deriv] -= 1
                coef_grad[d_deriv, :, i_coef] *= pow_grad[d_deriv, i_coef, d_deriv] + 1
            else:
                coef_grad[d_deriv, :, i_coef] = 0.0

    return pow_grad, coef_grad


def compute_hess_diag_poly_params(
    pow_arr: np.ndarray, coeff_arr: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """computes the power array and the coefficient array of the diagonal of the hessian of the multivariate polynomial defined with the given power and coefficient arrays

    Parameters
    ----------
    pow_arr : np.ndarray of shape (n_coefs, D_no_kappa)
        array of powers of the initial multivariate polynomial
    coeff_arr : np.ndarray of shape (L, n_coefs)
        array of coefficients of the initial multivariate polynomial

    Returns
    -------
    pow_grad : np.ndarray of shape (D_no_kappa, n_coefs, D_no_kappa)
        array of powers of the diagonal hessian of multivariate polynomial
    coef_grad : np.ndarray of shape (D_no_kappa, L, n_coefs)
        array of coefficients of the diagonal hessian of multivariate polynomial
    """
    L, n_coefs = coeff_arr.shape
    n_coefs_2, D_no_kappa = pow_arr.shape
    assert n_coefs == n_coefs_2

    pow_hess_diag = pow_arr[None, :, :] * np.ones((D_no_kappa, n_coefs, D_no_kappa))
    coef_hess_diag = coeff_arr[None, :, :] * np.ones((D_no_kappa, L, n_coefs))

    for d_deriv in range(D_no_kappa):
        for i_coef in range(n_coefs):
            if pow_hess_diag[d_deriv, i_coef, d_deriv] >= 2:
                pow_hess_diag[d_deriv, i_coef, d_deriv] -= 2
                coef_hess_diag[d_deriv, :, i_coef] *= (
                    pow_hess_diag[d_deriv, i_coef, d_deriv] + 1
                ) * (pow_hess_diag[d_deriv, i_coef, d_deriv] + 2)
            else:
                coef_hess_diag[d_deriv, :, i_coef] = 0.0

    return pow_hess_diag, coef_hess_diag


@nb.jit(nopython=True, fastmath=True)
def evaluate_poly(
    Theta: np.ndarray, pow_arr: np.ndarray, coeff_arr: np.ndarray, deg: int
) -> np.ndarray:
    N_pix, D_no_kappa = Theta.shape
    L, n_coefs = coeff_arr.shape

    Theta_pow = np.ones((N_pix, D_no_kappa, deg + 1))
    for pow_ in range(1, deg + 1):
        Theta_pow[:, :, pow_] = Theta_pow[:, :, pow_ - 1] * Theta

    f_Theta = np.zeros((N_pix, L))
    for idx in range(n_coefs):
        prod_ = np.ones((N_pix,))
        for d in range(D_no_kappa):
            prod_ *= Theta_pow[:, d, pow_arr[idx, d]]
        f_Theta[:, :] += np.expand_dims(coeff_arr[:, idx], 0) * np.expand_dims(prod_, 1)

    return f_Theta


@nb.jit(nopython=True, fastmath=True)
def grad_poly(
    Theta: np.ndarray, pow_grad: np.ndarray, coef_grad: np.ndarray, deg: int
) -> np.ndarray:
    N_pix, D_no_kappa = Theta.shape
    _, L, n_coefs = coef_grad.shape

    Theta_pow = np.ones((N_pix, D_no_kappa, deg + 1))
    for pow_ in range(1, deg + 1):
        Theta_pow[:, :, pow_] = Theta_pow[:, :, pow_ - 1] * Theta

    grad_f_Theta = np.zeros((N_pix, D_no_kappa, L))
    for d in range(D_no_kappa):
        for idx in range(n_coefs):
            prod_ = np.ones((N_pix,))
            for i in range(D_no_kappa):
                prod_ *= Theta_pow[:, i, pow_grad[d, idx, i]]
            grad_f_Theta[:, d, :] += np.expand_dims(
                coef_grad[d, :, idx], 0
            ) * np.expand_dims(prod_, 1)

    return grad_f_Theta


@nb.jit(nopython=True, fastmath=True)
def hess_diag_poly(
    Theta: np.ndarray, pow_hess_diag: np.ndarray, coef_hess_diag: np.ndarray, deg: int
) -> np.ndarray:
    N_pix, D_no_kappa = Theta.shape
    _, L, n_coefs = coef_hess_diag.shape

    Theta_pow = np.ones((N_pix, D_no_kappa, deg + 1))
    for pow_ in range(1, deg + 1):
        Theta_pow[:, :, pow_] = Theta_pow[:, :, pow_ - 1] * Theta

    hess_diag_f_Theta = np.zeros((N_pix, D_no_kappa, L))
    for d in range(D_no_kappa):
        for idx in range(n_coefs):
            prod_ = np.ones((N_pix,))
            for i in range(D_no_kappa):
                prod_ *= Theta_pow[:, i, pow_hess_diag[d, idx, i]]
            hess_diag_f_Theta[:, d, :] += np.expand_dims(
                coef_hess_diag[d, :, idx], 0
            ) * np.expand_dims(prod_, 1)

    return hess_diag_f_Theta


class PolynomialApprox(ExpForwardMap):
    r"""multi-line polynomial regressor made to approximate the true forward model given a grid of simulations.

    Given the fact that the true Meudon PDR code varies over orders of magnitudes and in order to guarantee that the interpolations are strictly positive, the regression is done in the log space. Therefore, for each line, the approximation is of the form

    .. math::

            f_{\ell} (\theta_n) = \exp \circ \ln f_{\ell} (\theta_n), \quad \ln f_{\ell} \in \mathbb{R}[X]
    """

    LOGE_10 = np.log(10.0)
    r"""float: natural log (in base :math:`e`) of 10, computed once and saved to limit redundant computations"""

    SIMU_PATH = os.path.abspath(f"{os.path.dirname(os.path.abspath(__file__))}/../../../data/")
    r"""path to the folder containing the data of saved models"""

    def __init__(
        self,
        model_name: str,
        dict_fixed_values_scaled: dict[str, Optional[float]],
        angle: float,
    ):
        filepath = f"{self.SIMU_PATH}/models/{model_name}/model.pickle"
        assert filepath[-7:] == ".pickle", "incorrect format"

        with open(filepath, "rb") as file_:
            (
                D_no_kappa,
                D,
                L,
                deg,
                pow_arr,
                coeff_arr,
            ) = pickle.load(file_)

        self.D_no_kappa = D_no_kappa
        r"""int: full dimension of the physical parameter vector, except for the scaling parameter :math:`\kappa` (which is not an input of the polynomial)"""

        self.D = D
        r"""int: full dimension of the physical parameter vector, including the scaling parameter :math:`\kappa`"""

        self.L = L
        r"""int: total number of observables per pixel used for inversion"""

        self.deg = deg
        r"""int: degree of the polynomial"""

        self.set_sampled_and_fixed_entries(dict_fixed_values_scaled)

        self.pow_arr = pow_arr  # (n_coefs, D_no_kappa)
        r"""np.ndarray of shape (n_coefs, D_no_kappa): powers of the monomials (loaded from model)"""
        self.coeff_arr = coeff_arr  # (L, n_coefs)
        r"""np.ndarray of shape (L, n_coefs): coefficients associated to the monomials (loaded from model)"""

        self.angle = angle
        r"""float: observation angle"""

        pow_grad, coeff_grad = compute_grad_poly_params(
            self.pow_arr,
            self.coeff_arr,
        )
        self.pow_grad = pow_grad.astype(int)  # (D_no_kappa, n_coefs, D_no_kappa)
        r"""np.ndarray of shape (D_no_kappa, n_coefs, D_no_kappa): powers of the monomials of the first derivative polynomial (computed in ``__init__`` method)"""
        self.coeff_grad = coeff_grad.astype(float)  # (D_no_kappa, L, n_coefs)
        r"""np.ndarray of shape (D_no_kappa, L, n_coefs): coefficients associated to the monomials of the first derivative polynomial (computed in ``__init__`` method)"""

        pow_hess_diag, coeff_hess_diag = compute_hess_diag_poly_params(
            self.pow_arr, self.coeff_arr
        )
        self.pow_hess_diag = pow_hess_diag.astype(int)
        r"""np.ndarray of shape (D_no_kappa, n_coefs, D_no_kappa): powers of the monomials of the diagonal terms of the second derivative polynomial (computed in ``__init__`` method)"""
        self.coeff_hess_diag = coeff_hess_diag.astype(float)
        r"""np.ndarray of shape (D_no_kappa, L, n_coefs): coefficients associated to the monomials of the first derivative polynomial (computed in ``__init__`` method)"""

        with open(
            f"{self.SIMU_PATH}/models/{model_name}/line_names.pickle",
            "rb",
        ) as f:
            self.outputs_names = pickle.load(f)
            r"""list: names of the observables"""

        self.output_subset = self.outputs_names * 1
        r"""list: names of the observables used for inversion"""
        self.output_subset_indices = list(range(L))
        r"""list: indices the observables"""

        self.coeff_arr_subset = self.coeff_arr[self.output_subset_indices, :] * 1
        r"""np.ndarray: coefficients of the polynomial with restricted outputs"""
        self.coeff_grad_subset = self.coeff_grad[:, self.output_subset_indices, :] * 1
        r"""np.ndarray: coefficients of the first derivative polynomial with restricted outputs"""
        self.coeff_hess_diag_subset = (
            self.coeff_hess_diag[:, self.output_subset_indices, :] * 1
        )
        r"""np.ndarray: coefficients of the diagonal of the second derivative polynomial with restricted outputs"""

    @staticmethod
    def _check_if_subsequence(seq: Sequence, subseq: Sequence) -> bool:
        return set(subseq) <= set(seq)

    @staticmethod
    def _indices_of_subsequence(seq: Sequence, subseq: Sequence) -> List[int]:
        index_dict = dict((value, idx) for idx, value in enumerate(seq))
        return [index_dict[value] for value in subseq]  # Remark: the result is ordered

    def indices_output_subset(
        self, output_subset: Union[Sequence[str], Sequence[int]]
    ) -> List[int]:
        if len(output_subset) == 0:
            raise ValueError("output_subset must not be empty")
        if isinstance(output_subset[0], str):
            if self.outputs_names is None:
                raise TypeError(
                    "output_subset cannot be a sequence of str when self.outputs_names is None"
                )
            if not self._check_if_subsequence(self.outputs_names, output_subset):
                raise ValueError("output_subset is not a valid subset")
            indices = indices = self._indices_of_subsequence(
                self.outputs_names, output_subset
            )
        elif isinstance(output_subset[0], int):
            if not self._check_if_subsequence(
                list(range(self.out_features)), output_subset
            ):
                raise ValueError("input_subset is not a valid subset")
            indices = output_subset
        else:
            raise TypeError(
                f"output_subset must contain str or int, not {type(output_subset[0])}"
            )

        return indices

    def restrict_to_output_subset(self, output_subset: Sequence[str]) -> None:
        for line in output_subset:
            assert line in self.outputs_names, line

        self.output_subset = output_subset * 1
        self.output_subset_indices = self.indices_output_subset(output_subset)
        self.L = len(output_subset)

        self.coeff_arr_subset = self.coeff_arr[self.output_subset_indices, :] * 1
        self.coeff_grad_subset = self.coeff_grad[:, self.output_subset_indices, :] * 1
        self.coeff_hess_diag_subset = (
            self.coeff_hess_diag[:, self.output_subset_indices, :] * 1
        )

    def evaluate(self, Theta: np.ndarray) -> np.ndarray:
        Theta_with_angle = np.column_stack(
            (Theta, self.angle * np.ones((Theta.shape[0], 1))),
        )
        val = evaluate_poly(
            Theta_with_angle[:, 1:],
            self.pow_arr,
            self.coeff_arr_subset,
            self.deg,
        )
        val += Theta[:, 0][:, None]  # add log_kappa
        return np.exp(val)  # (N, L)

    def evaluate_log(self, Theta: np.ndarray) -> np.ndarray:
        Theta_with_angle = np.column_stack(
            (Theta, self.angle * np.ones((Theta.shape[0], 1))),
        )
        val = evaluate_poly(
            Theta_with_angle[:, 1:],
            self.pow_arr,
            self.coeff_arr_subset,
            self.deg,
        )
        val += Theta[:, 0][:, None]  # add log_kappa
        return val  # (N, L)

    def gradient(self, Theta: np.ndarray) -> np.ndarray:
        Theta_with_angle = np.column_stack(
            (Theta, self.angle * np.ones((Theta.shape[0], 1))),
        )
        theta = Theta_with_angle[:, 1:]
        grad_P = self._gradient_log(theta)  # (N, D, L)
        Intensities = self.evaluate(Theta)  # (N, L)
        return grad_P * Intensities[:, None, :]  # (N, D, L)

    def gradient_log(self, Theta: np.ndarray) -> np.ndarray:
        Theta_with_angle = np.column_stack(
            (Theta, self.angle * np.ones((Theta.shape[0], 1))),
        )
        theta = Theta_with_angle[:, 1:]
        return self._gradient_log(theta)  # (N, D, L)

    def _gradient_log(self, theta):
        grad_ = grad_poly(
            theta,
            self.pow_grad,
            self.coeff_grad_subset,
            self.deg,
        )
        grad_ = grad_[:, :-1, :]  # (N, D_no_kappa, L) (remove angle)

        # gradient that includes kappa (the grad of p wrt kappa is 1)
        grad_full = np.ones((grad_.shape[0], self.D, self.L))
        grad_full[:, 1:, :] = grad_
        return grad_full  # (N, D, L)

    def _hess_diag_log(self, theta):
        hess_diag = hess_diag_poly(
            theta, self.pow_hess_diag, self.coeff_hess_diag_subset, self.deg
        )
        hess_diag = hess_diag[:, :-1, :]  # (N, D_no_kappa, L) (remove angle)

        hess_diag_full = np.zeros((theta.shape[0], self.D, self.L))
        hess_diag_full[:, 1:, :] = hess_diag  # (N, D, L)

        return hess_diag_full  # (N, D, L)

    def hess_diag(self, Theta: np.ndarray) -> np.ndarray:
        Theta_with_angle = np.column_stack(
            (Theta, self.angle * np.ones((Theta.shape[0], 1))),
        )
        theta = Theta_with_angle[:, 1:]

        diag_hess_log = self._hess_diag_log(theta)
        grad_log = self._gradient_log(theta)
        Intensities = self.evaluate(Theta)

        diag_hess = (diag_hess_log + grad_log**2) * Intensities[:, None, :]
        return diag_hess  # (N, D, L)

    def hess_diag_log(self, Theta: np.ndarray) -> np.ndarray:
        Theta_with_angle = np.column_stack(
            (Theta, self.angle * np.ones((Theta.shape[0], 1))),
        )
        theta = Theta_with_angle[:, 1:]
        return self._hess_diag_log(theta)  # (N, D, L)

    def compute_all(
        self,
        Theta: np.ndarray,
        compute_lin: bool = True,
        compute_log: bool = True,
        compute_derivatives: bool = True,
        compute_derivatives_2nd_order: bool = True,
    ) -> dict:
        r"""gathers the evaluation of the forward map in linear and log scale and of the associated derivatives. Permits to limit repeating computations, but requires the storage in memory of the result.

        Parameters
        ----------
        Theta : np.ndarray of shape (N, D)
            array of points in the input space :math:`\Theta = (\theta_n)_{n=1}^N` with :math:`\theta_n \in \mathbb{R}^D`
        compute_lin : bool, optional
            wether or not to compute the forward model (and possibly the gradient and diagonal of the Hessian), by default True
        compute_log : bool, optional
            wether or not to compute the log-forward model (and possibly the gradient and diagonal of the Hessian), by default True
        compute_derivatives : bool, optional
            wether or not to evaluate the derivatives of the forward map, by default True
        compute_derivatives_2nd_order : bool, optional
            wether or not to evaluate the 2nd order derivatives of the forward map, by default True

        Returns
        -------
        forward_map_evals : dict[str, np.ndarray]
            dictionary with entries such as `f_Theta`, `log_f_Theta`, `grad_f_Theta`, `grad_log_f_Theta`, `hess_diag_f_Theta` and `hess_diag_log_f_Theta`, depending on the input booleans.

        Note
        ----
        To evaluating :math:`f(\theta_n)` and the associated derivatives, 3 evaluations are enough for six functions. Calling each function would result in a total of 9 evaluations."""
        forward_map_evals = dict()

        N_pix = Theta.shape[0]

        if compute_derivatives:
            log_f_Theta = self.evaluate_log(Theta)
            grad_log_f_Theta = self.gradient_log(Theta)

            log_f_Theta *= self.LOGE_10
            grad_log_f_Theta = grad_log_f_Theta[:, : self.D_no_kappa, :] * self.LOGE_10

            if compute_derivatives_2nd_order:
                hess_diag_log_f_Theta = self.hess_diag(Theta)
                hess_diag_log_f_Theta = (
                    hess_diag_log_f_Theta[:, : self.D_no_kappa, :] * self.LOGE_10
                )

            log_f_Theta = Theta[:, 0][:, None] + log_f_Theta

            if compute_log:
                forward_map_evals["log_f_Theta"] = log_f_Theta

                grad_log_f_Theta_full = np.ones((N_pix, self.D, self.L))
                grad_log_f_Theta_full[:, 1:, :] = grad_log_f_Theta * 1
                forward_map_evals["grad_log_f_Theta"] = grad_log_f_Theta_full

                if compute_derivatives_2nd_order:
                    hess_diag_log_f_Theta_full = np.zeros((N_pix, self.D, self.L))
                    hess_diag_log_f_Theta_full[:, 1:, :] = hess_diag_log_f_Theta * 1
                    forward_map_evals[
                        "hess_diag_log_f_Theta"
                    ] = hess_diag_log_f_Theta_full

            if compute_lin:
                f_Theta = np.exp(log_f_Theta)
                forward_map_evals["f_Theta"] = f_Theta

                # (N_pix, D, L)
                forward_map_evals["grad_f_Theta"] = (
                    grad_log_f_Theta_full * f_Theta[:, None, :]
                )

                if compute_derivatives_2nd_order:
                    # (N_pix, D, L)
                    forward_map_evals["hess_diag_f_Theta"] = f_Theta[:, None, :] * (
                        hess_diag_log_f_Theta_full + grad_log_f_Theta_full**2
                    )

            return forward_map_evals

        else:
            log_f_Theta = self.evaluate_log(Theta)
            if compute_log:
                forward_map_evals["log_f_Theta"] = log_f_Theta
            if compute_lin:
                forward_map_evals["f_Theta"] = np.exp(log_f_Theta)
            return forward_map_evals



import json
from functools import partial
from typing import Dict, List

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from bayes_opt import BayesianOptimization
from bayes_opt.event import Events
from bayes_opt.logger import JSONLogger
from matplotlib.ticker import FuncFormatter
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Matern
from tqdm.auto import tqdm

import beetroots.approx_optim.approach_type.utils as utils
from beetroots.approx_optim.approach_type.abstract_approach_type import ApproachType


class BayesianOptimizationApproach(ApproachType):
    r"""implements a Bayesian optimization approach to adjust the likelihood parameter"""

    def optimization(
        self,
        first_points: List,
        init_points: int,
        n_iter: int,
        list_log10_f_grid: np.ndarray,
        pdf_kde_log10_f_Theta: np.ndarray,
        pbounds: Dict,
        sigma_a_val: float,
        sigma_m_val: float,
        n: int,
        ell: int,
    ) -> None:
        """performs Bayesian optimization

        Parameters
        ----------
        first_points : List
            list of the points to evaluate before running the optimization
        init_points : int
            number of init points
        n_iter : int
            number of optimization steps to perform
        list_log10_f_grid : np.ndarray
            _description_
        pdf_kde_log10_f_Theta : np.ndarray
            _description_
        pbounds : Dict
            _description_
        sigma_a_val : float
            _description_
        sigma_m_val : float
            _description_
        n : int
            pixel index for which the optimization is performed
        ell : int
            line index for which the optimization is performed
        """

        # fix secondary parameters
        estimate_avg_dks = partial(
            utils.estimate_avg_dks_full_bo,
            list_log10_f_grid=list_log10_f_grid,
            pdf_kde_log10_f_Theta=pdf_kde_log10_f_Theta,
            sigma_a=sigma_a_val,
            sigma_m=sigma_m_val,
            N_samples_y=self.N_samples_y,
            max_workers=self.max_workers,
        )

        # setup optimization
        optimizer = BayesianOptimization(
            f=estimate_avg_dks,
            pbounds=pbounds,
            verbose=2,
            random_state=1,
        )

        logger = JSONLogger(
            path=f"{self.path_logs}/logs_n{n}_{self.list_lines[ell]}.json"
        )
        optimizer.subscribe(Events.OPTIMIZATION_STEP, logger)

        # preliminary points
        for point in first_points:
            optimizer.probe(
                params={"a0": point[0], "a1": point[1]},
                lazy=True,
            )

        # perform optimization
        optimizer.maximize(init_points=init_points, n_iter=n_iter)
        return

    def plots_postprocessing(
        self,
        bounds_a0_low: np.ndarray,
        bounds_a0_high: np.ndarray,
        bounds_a1_low: float,
        bounds_a1_high: float,
        n_iter: int,
    ) -> None:
        """makes some plots after the optimization process to facilitate the execution understanding

        Parameters
        ----------
        bounds_a0_low : np.ndarray
            array of lower bounds on the transition location
        bounds_a0_high : np.ndarray
            array of upper bounds on the transition location
        bounds_a1_low : float
            lower bound on the transition slope
        bounds_a1_high : float
            upper bound on the transition slope
        n_iter : int
            number of performed optimization iterations
        """
        self.plot_evaluation_points(
            bounds_a0_low,
            bounds_a0_high,
            bounds_a1_low,
            bounds_a1_high,
        )
        self.plot_GP(
            bounds_a0_low,
            bounds_a0_high,
            bounds_a1_low,
            bounds_a1_high,
            n_iter,
        )
        return

    def plot_evaluation_points(
        self,
        bounds_a0_low: np.ndarray,
        bounds_a0_high: np.ndarray,
        bounds_a1_low: float,
        bounds_a1_high: float,
    ) -> None:
        """plots the sequence of sampled points with the order as colorbar

        Parameters
        ----------
        bounds_a0_low : np.ndarray
            array of lower bounds on the transition location
        bounds_a0_high : np.ndarray
            array of upper bounds on the transition location
        bounds_a1_low : float
            lower bound on the transition slope
        bounds_a1_high : float
            upper bound on the transition slope
        """
        for n in range(self.N):
            for ell in tqdm(range(self.L)):
                with open(
                    f"{self.path_logs}/logs_correct_format_n{n}_{self.list_lines[ell]}.json",
                    "r",
                ) as f:
                    data = json.load(f)

                # Flatten data
                df = pd.json_normalize(data, max_level=1)
                df = df.rename(columns={"params.a0": "a0", "params.a1": "a1"})

                plt.figure(figsize=(8, 6))
                plt.title("order of evaluation")
                plt.scatter(df["a0"], df["a1"], c=df.index.values)
                plt.xlim([bounds_a0_low[n, ell] - 0.5, bounds_a0_high[n, ell] + 0.5])
                plt.ylim([bounds_a1_low - 0.1, bounds_a1_high + 0.1])
                plt.colorbar()
                plt.grid()
                plt.savefig(
                    f"{self.path_img_final}/order_points_n{n}_{self.list_lines[ell]}.PNG"
                )
                plt.close()

                plt.figure(figsize=(8, 6))
                plt.title("target evaluated values")
                plt.scatter(df["a0"], df["a1"], c=df["target"])
                plt.xlim([bounds_a0_low[n, ell] - 0.5, bounds_a0_high[n, ell] + 0.5])
                plt.ylim([bounds_a1_low - 0.1, bounds_a1_high + 0.1])
                plt.colorbar()
                plt.grid()
                plt.savefig(
                    f"{self.path_img_final}/target_points_n{n}_{self.list_lines[ell]}.PNG"
                )
                plt.close()

        return

    def plot_GP(
        self,
        bounds_a0_low: np.ndarray,
        bounds_a0_high: np.ndarray,
        bounds_a1_low: float,
        bounds_a1_high: float,
        n_iter: int,
    ) -> None:
        """plot the state (both mean and standard deviations) of the Gaussian process at multiple steps of the optimization process.

        Parameters
        ----------
        bounds_a0_low : np.ndarray
            array of lower bounds on the transition location
        bounds_a0_high : np.ndarray
            array of upper bounds on the transition location
        bounds_a1_low : float
            lower bound on the transition slope
        bounds_a1_high : float
            upper bound on the transition slope
        n_iter : int
            number of performed optimization iterations
        """
        # show_theta_scale = True
        # show_y_scale = True

        gp = GaussianProcessRegressor(
            kernel=Matern(nu=2.5),
            alpha=1e-2,
            normalize_y=True,
            n_restarts_optimizer=5,
            random_state=1,
        )
        n_a0 = 100
        n_a1 = 100

        fmt_mu = lambda x, pos: "{:.2f}".format(x)
        fmt_sigma = lambda x, pos: "{:.3f}".format(x)

        for n in range(self.N):
            for ell in tqdm(range(self.L)):
                with open(
                    f"{self.path_logs}/logs_correct_format_n{n}_{self.list_lines[ell]}.json",
                    "r",
                ) as f:
                    data = json.load(f)

                # Flatten data
                df = pd.json_normalize(data, max_level=1)
                df = df.rename(columns={"params.a0": "a0", "params.a1": "a1"})

                list_a0 = np.linspace(
                    bounds_a0_low[n, ell], bounds_a0_high[n, ell], n_a0
                )
                list_a1 = np.linspace(bounds_a1_low, bounds_a1_high, n_a1)
                AA0, AA1 = np.meshgrid(list_a0, list_a1)

                assert AA0.shape == (n_a1, n_a0)

                grid = np.vstack((AA0.flatten(), AA1.flatten())).T  # (n_a0 * n_a1, 2)

                sigma_a = self.sigma_a[n, ell]
                sigma_m = self.sigma_m[n, ell]

                var_eps_m = np.exp(sigma_m**2) * (np.exp(sigma_m**2) - 1)
                log10_f0 = 0.5 * (2 * np.log(sigma_a) - np.log(var_eps_m)) / np.log(10)

                # with full dataset
                a0_obs_full = df.loc[:, "a0"].values
                a1_obs_full = df.loc[:, "a1"].values
                x_obs_full = np.vstack((a0_obs_full, a1_obs_full)).T
                y_obs_full = df.loc[:, "target"].values

                df_sorted = df.sort_values("target", ascending=False)
                df_sorted = df_sorted.reset_index()
                a0_best, a1_best = df_sorted.loc[0, ["a0", "a1"]]

                gp.fit(x_obs_full, y_obs_full)
                mu_full, sigma_full = gp.predict(grid, return_std=True)

                mu_full = mu_full.reshape((n_a1, n_a0))
                sigma_full = sigma_full.reshape((n_a1, n_a0))

                mu_min = mu_full.min()
                mu_max = mu_full.max()
                sigma_min = sigma_full.min()
                sigma_max = sigma_full.max()

                for i_iter, show_estimate in zip([5, 20, n_iter], [False, False, True]):
                    a0_obs = df.loc[:i_iter, "a0"].values
                    a1_obs = df.loc[:i_iter, "a1"].values
                    x_obs = np.vstack((a0_obs, a1_obs)).T
                    y_obs = df.loc[:i_iter, "target"].values

                    gp.fit(x_obs, y_obs)
                    mu, sigma = gp.predict(grid, return_std=True)
                    mu = mu.reshape((n_a1, n_a0))
                    sigma = sigma.reshape((n_a1, n_a0))

                    plt.figure(figsize=(8, 6))
                    plt.title(f"GP mean after {i_iter} iterations", fontsize=24)
                    plt.contourf(
                        AA0,
                        AA1,
                        mu,
                        levels=100,
                        vmax=mu_max,
                        vmin=mu_min,
                    )
                    plt.axvline(log10_f0, c="r", ls="--")  # , label="equal variances")
                    plt.colorbar(format=FuncFormatter(fmt_mu))
                    plt.scatter(a0_obs, a1_obs, c="k")

                    if show_estimate:
                        plt.scatter(
                            [a0_best],
                            [a1_best],
                            c="r",
                            marker="+",
                            label=r"$\hat{a}$",
                            linewidths=3.5,
                            s=225,
                        )
                        plt.legend()

                    # if show_theta_scale:
                    plt.xlabel(r"interval center $(a_1 + a_0) \; / \; 2$")
                    # plt.xticks([-10, -8, -6, -4, -2], [-10, -8, -6, -4, -2])
                    # else:
                    # plt.xticks([-10, -8, -6, -4, -2], [])

                    # if show_y_scale:
                    plt.ylabel(r"interval radius $(a_1 - a_0) \; / \; 2$")
                    # plt.yticks([0.5, 1.0, 1.5, 2.0], [0.5, 1.0, 1.5, 2.0])
                    # else:
                    # plt.yticks([0.5, 1.0, 1.5, 2.0], [])

                    plt.tight_layout()
                    plt.savefig(
                        f"{self.path_img_final}/gp_n{n}_{self.list_lines[ell]}_mean_iter{i_iter}.PNG",
                        bbox_inches="tight",
                        transparent=True,
                    )
                    plt.close()

                    plt.figure(figsize=(8, 6))
                    plt.title(f"GP std after {i_iter} iterations", fontsize=24)
                    # plt.title("std of GP")
                    plt.contourf(
                        AA0,
                        AA1,
                        sigma,
                        levels=100,
                        vmax=sigma_max,
                        vmin=sigma_min,
                    )
                    plt.axvline(log10_f0, c="r", ls="--")  # , label="equal variances")
                    plt.colorbar(format=FuncFormatter(fmt_sigma))
                    plt.scatter(a0_obs, a1_obs, c="k")

                    if show_estimate:
                        plt.scatter(
                            [a0_best],
                            [a1_best],
                            c="r",
                            marker="+",
                            label=r"$\hat{a}$",
                            linewidths=3.5,
                            s=225,
                        )
                        plt.legend()

                    # if show_theta_scale:
                    plt.xlabel(r"interval center $(a_1 + a_0) \; / \; 2$")
                    #     plt.xticks([-10, -8, -6, -4, -2], [-10, -8, -6, -4, -2])
                    # else:
                    #     plt.xticks([-10, -8, -6, -4, -2], [])

                    # if show_y_scale:
                    plt.ylabel(r"interval radius $(a_1 - a_0) \; / \; 2$")
                    #     plt.yticks([0.5, 1.0, 1.5, 2.0], [0.5, 1.0, 1.5, 2.0])
                    # else:
                    #     plt.yticks([0.5, 1.0, 1.5, 2.0], [])

                    # plt.ylabel(r"mixing interval radius $(a_1 - a_0)/2$")
                    plt.tight_layout()

                    plt.savefig(
                        f"{self.path_img_final}/gp_n{n}_{self.list_lines[ell]}_std_iter{i_iter}.PNG",
                        bbox_inches="tight",
                        transparent=True,
                    )
                    plt.close()

        return

    def plot_cost_map(self):
        pass



import abc
import os
from typing import List, Optional

import h5py
import numpy as np

from beetroots.space_transform.abstract_transform import Scaler


class Saver:
    r"""enable to regularly save the progression of the Markov chain to a ``.hdf5`` file
    """

    __slots__ = (
        "N",
        "D",
        "D_sampling",
        "L",
        "results_path",
        "batch_size",
        "freq_save",
        "scaler",
        "t_last_save",
        "t_last_init",
        "next_batch_size",
        "final_next_batch_size",
        "memory",
        "save_forward_map_evals",
        "list_idx_sampling",
    )

    def __init__(
        self,
        N: int,
        D: int,
        D_sampling: int,
        L: int,
        scaler: Scaler,
        results_path: str = "",
        batch_size: Optional[int] = None,
        freq_save: int = 1,
        save_forward_map_evals: bool = False,
        list_idx_sampling: Optional[List[int]] = None,
    ):
        """

        Parameters
        ----------
        N : int
            total number of pixels to reconstruct
        D : int
            total number of physical parameters
        D_sampling : int
            number of physical parameters that are optimized / sampled
        L : int
            number of observed lines
        scaler : Scaler
            contains the transformation of the Theta values from their scaled space (in which the sampling happens) to their natural space
        results_path : str
            path towards the ``.hdf5`` output file, by default ""
        batch_size : int, optional
            number of iterations between two saves on file, by default None
        freq_save : int, optional
            save one sample in every (freq_save). Used to save disk space., by default 1
        save_forward_map_evals: bool, optional
            wether to save the forward model evaluations and gradients, by default False
        list_idx_sampling : Optional[List[int]], optional
            contains the indices of the physical parameters to be sampled
        """
        self.N = N
        r"""int: total number of pixels to reconstruct"""
        self.D = D
        r"""int: total number of physical parameters"""
        self.D_sampling = D_sampling
        r"""int: number of physical parameters that are optimized / sampled"""
        self.L = L
        r"""int: number of observed lines"""

        if list_idx_sampling is None:
            list_idx_sampling = np.arange(self.D)

        self.list_idx_sampling = list_idx_sampling
        r"""1D np.ndarray: contains the indices of the physical parameters to be sampled"""

        self.results_path = results_path
        r"""str: path towards the ``.hdf5`` output file"""
        if len(results_path) > 0 and not (os.path.isdir(results_path)):
            os.mkdir(results_path)

        self.batch_size = batch_size
        r"""int: frequency of saves, i.e., "every ``batch_size`` new iterates to be saved, the memory is saved to an ``.hdf5`` file and re-initialized"""

        self.freq_save = freq_save
        r"""int: frequency of saved iterates during the run (1 means that every iteration is saved)"""

        self.scaler = scaler
        r"""Scaler: contains the transformation of the Theta values from their natural space to their scaled space (in which the sampling happens)"""

        # these two attributes are initialized by initialize_memory
        # updated during sampling
        self.t_last_save = 0
        r"""int: time index of the last save of the memory to ``.hdf5`` file"""
        self.t_last_init = 0
        r"""int: time index of the last memory initialization"""
        self.next_batch_size = 0
        r"""int: number of iterates to be stored in the next batch, i.e., until next save to file"""
        self.final_next_batch_size = 0
        r"""int: """
        self.memory = dict()
        """dict[str, Union[float, np.ndarray]]: stores the values before saving them to file"""

        self.save_forward_map_evals = save_forward_map_evals
        r"""bool: wether to save the forward model evaluations and gradients"""

    def set_results_path(self, results_path: str) -> None:
        r"""sets the path of the ``.hdf5`` file to a new value

        Parameters
        ----------
        results_path : str
            path towards the ``.hdf5`` output file
        """
        self.results_path = results_path
        if len(results_path) > 0 and not (os.path.isdir(results_path)):
            os.mkdir(results_path)
        return

    def check_need_to_save(self, t: int) -> bool:
        """checks wether or not the memory should be saved to a ``.hdf5`` file

        Parameters
        ----------
        t : int
            current iteration index

        Returns
        -------
        bool
            wether or not to save to disk now
        """
        current_t_in_batch = t - self.t_last_init + 1
        return current_t_in_batch == self.next_batch_size * self.freq_save

    def check_need_to_update_memory(self, t: int) -> bool:
        """checks wether or not the memory should be updated

        Parameters
        ----------
        t : int
            current iteration index

        Returns
        -------
        bool
            wether or not to save to update the memory
        """
        return t % self.freq_save == 0

    @abc.abstractmethod
    def initialize_memory(
        self,
        T_MC: int,
        t: int,
        Theta: np.ndarray,
        forward_map_evals: dict = dict(),
        nll_utils: dict = dict(),
        dict_objective: dict = dict(),
        additional_sampling_log: dict = dict(),
    ) -> None:
        r"""initializes the memory with the correct shapes"""
        pass

    @abc.abstractmethod
    def update_memory(
        self,
        t: int,
        Theta: np.ndarray,
        forward_map_evals: dict = dict(),
        nll_utils: dict = dict(),
        dict_objective: dict = dict(),
        additional_sampling_log: dict = dict(),
        rng_state_array: Optional[np.ndarray] = None,
        rng_inc_array: Optional[np.ndarray] = None,
    ) -> None:
        """updates the memory with new information. All of the potential entries are optional except for the current iterate."""
        pass

    def save_to_file(self):
        """Saves the current memory content to a ``.hdf5`` file"""

        if self.t_last_init == 1:  # if first writing
            with h5py.File(
                os.path.join(self.results_path, "mc_chains.hdf5"),
                "w",
            ) as f:
                for k, v in self.memory.items():
                    f.create_dataset(k, data=v, maxshape=(None,) + v.shape[1:])

        else:  # append data to already created file
            with h5py.File(
                os.path.join(self.results_path, "mc_chains.hdf5"),
                "a",
            ) as f:
                for k, v in self.memory.items():
                    f[k].resize(
                        f[k].shape[0] + self.final_next_batch_size,
                        axis=0,
                    )
                    f[k][-self.final_next_batch_size :] = v

        self.memory = dict()

    def save_additional(
        self,
        list_arrays: List[np.ndarray],
        list_names: List[str],
    ) -> None:
        r"""saves additional content to a ``.hdf5`` file

        Parameters
        ----------
        list_arrays : List[np.ndarray]
            list of the arrays to be saved
        list_names : List[str]
            list of names for the arrays to be saved in the ``.hdf5`` file
        """
        assert len(list_names) == len(list_arrays)

        with h5py.File(
            os.path.join(self.results_path, "mc_chains.hdf5"),
            "a",
        ) as f:
            for name, array_ in zip(list_names, list_arrays):
                f.create_dataset(name, data=array_)

        return



import os
from typing import Optional, Tuple

import numpy as np
import pandas as pd


class EstimatorPerfSaver:

    __slots__ = ()

    def __init__(self):
        pass

    @staticmethod
    def compute_MSE(Theta_estimate: np.ndarray, Theta_true: np.ndarray):
        return np.linalg.norm(Theta_true - Theta_estimate) ** 2

    @staticmethod
    def compute_SNR(Theta_estimate: np.ndarray, Theta_true: np.ndarray):
        denom = np.linalg.norm(Theta_true) ** 2
        mse_ = EstimatorPerfSaver.compute_MSE(Theta_estimate, Theta_true)
        return -10 * np.log10(mse_ / denom)

    @staticmethod
    def estimate_point_with_lowest_obj(
        list_Theta_lin: np.ndarray,
        list_objective: np.ndarray,
    ) -> Tuple[np.ndarray, float]:
        idx_best = np.argmin(list_objective)
        Theta_estimate = list_Theta_lin[idx_best]
        objective_estimate = list_objective[idx_best]
        return Theta_estimate, objective_estimate

    def save_estimator_performance(
        self,
        path_data_csv_out: str,
        estimator_name: str,
        model_name: str,
        mse: Optional[float],
        snr: Optional[float],
        objective: Optional[float],
    ) -> None:
        path_overall_results = f"{path_data_csv_out}/results_overall.csv"

        list_results_overall = [
            {
                "estimator": estimator_name,
                "model_name": model_name,
                "MSE": mse,
                "SNR": snr,
                "objective": objective,
            }
        ]

        df_results_overall = pd.DataFrame(list_results_overall)
        df_results_overall.to_csv(
            path_overall_results,
            mode="a",
            header=not (os.path.exists(path_overall_results)),
        )



import time
from typing import Dict, List, Optional, Union

import numpy as np

from beetroots.inversion.results.results_mcmc import ResultsExtractorMCMC
from beetroots.inversion.results.results_optim_map import ResultsExtractorOptimMAP
from beetroots.inversion.run.run_mcmc import RunMCMC
from beetroots.inversion.run.run_optim_map import RunOptimMAP
from beetroots.modelling.likelihoods.approx_censored_add_mult import (
    MixingModelsLikelihood,
)
from beetroots.modelling.likelihoods.gaussian_censored import CensoredGaussianLikelihood
from beetroots.modelling.posterior import Posterior
from beetroots.modelling.priors.l22_laplacian_prior import L22LaplacianSpatialPrior
from beetroots.modelling.priors.smooth_indicator_prior import SmoothIndicatorPrior
from beetroots.modelling.priors.spatial_prior_params import SpatialPriorParams
from beetroots.sampler.my_sampler import MySampler
from beetroots.sampler.saver.my_saver import MySaver
from beetroots.sampler.utils.my_sampler_params import MySamplerParams
from beetroots.simulations.astro.posterior_type.abstract_posterior_type import (
    SimulationPosteriorType,
)
from beetroots.space_transform.transform import MyScaler


class SimulationMySampler(SimulationPosteriorType):
    def setup_posteriors(
        self,
        scaler,
        forward_map,
        y,
        sigma_a,
        sigma_m,
        omega,
        syn_map,
        with_spatial_prior: bool,
        spatial_prior_params: SpatialPriorParams,
        indicator_margin_scale: float,
        lower_bounds_lin: Union[np.ndarray, List[float]],
        upper_bounds_lin: Union[np.ndarray, List[float]],
        list_gaussian_approx_params: List[bool],
        list_mixing_model_params: List,
    ) -> None:

        if with_spatial_prior:
            prior_spatial = L22LaplacianSpatialPrior(
                spatial_prior_params,
                self.cloud_name,
                N=self.N,
                D=self.D_sampling,
                df=syn_map,
                list_idx_sampling=self.list_idx_sampling,
            )
        else:
            prior_spatial = None

        # indicator prior
        if isinstance(lower_bounds_lin, list):
            lower_bounds_lin = np.array(lower_bounds_lin)
        if isinstance(upper_bounds_lin, list):
            upper_bounds_lin = np.array(upper_bounds_lin)

        lower_bounds = scaler.from_lin_to_scaled(
            lower_bounds_lin.reshape((1, self.D)),
        ).flatten()
        upper_bounds = scaler.from_lin_to_scaled(
            upper_bounds_lin.reshape((1, self.D)),
        ).flatten()
        prior_indicator = SmoothIndicatorPrior(
            self.D_sampling,
            self.N,
            indicator_margin_scale,
            lower_bounds,
            upper_bounds,
            list_idx_sampling=self.list_idx_sampling,
        )
        prior_indicator_1pix = SmoothIndicatorPrior(
            self.D_sampling,
            1,
            indicator_margin_scale,
            lower_bounds,
            upper_bounds,
            list_idx_sampling=self.list_idx_sampling,
        )

        # likelihood

        dict_posteriors = {}
        for i, dict_params in enumerate(list_mixing_model_params):
            model_name = f"mixing_{i}"

            likelihood_mixing = MixingModelsLikelihood(
                forward_map,
                self.D_sampling,
                self.L,
                self.N,
                y,
                sigma_a,
                sigma_m,
                omega,
                path_transition_params=dict_params["path_transition_params"],
                list_lines_fit=self.list_lines_fit * 1,
            )
            # separable+True: if no spatial prior, still possible to
            # run chromatic Gibbs on MTM
            posterior_mixing = Posterior(
                self.D_sampling,
                self.L,
                self.N,
                likelihood=likelihood_mixing,
                prior_spatial=prior_spatial,
                prior_indicator=prior_indicator,
                separable=True,  # only used if no spatial prior
            )
            dict_posteriors[model_name] = posterior_mixing

        for is_raw in list_gaussian_approx_params:
            name = "raw" if is_raw else "transformed"
            model_name = f"gaussian_approx_{name}"

            if is_raw:
                m_a = np.zeros((self.N, self.L))
                s_a = sigma_a * 1
            else:
                m_a = 0  # (np.exp(sigma_m**2 / 2) - 1) * y
                s_a = np.sqrt(
                    y**2
                    # * np.exp(sigma_m**2)
                    * (np.exp(sigma_m**2) - 1)
                    + sigma_a**2
                )

            likelihood_censor = CensoredGaussianLikelihood(
                forward_map,
                self.D_sampling,
                self.L,
                self.N,
                y,
                s_a,
                omega,
                bias=m_a,
            )
            # separable+True: if no spatial prior, still possible to run
            # chromatic Gibbs on MTM
            posterior_censor = Posterior(
                self.D_sampling,
                self.L,
                self.N,
                likelihood=likelihood_censor,
                prior_spatial=prior_spatial,
                prior_indicator=prior_indicator,
                separable=False,
            )
            dict_posteriors[model_name] = posterior_censor

        return dict_posteriors, scaler, prior_indicator_1pix

    def inversion_optim_mle(self):
        pass

    def inversion_optim_map(
        self,
        dict_posteriors: Dict[str, Posterior],
        scaler: MyScaler,
        my_sampler_params: MySamplerParams,
        N_MCMC: int,
        T_MC: int,
        T_BI: int,
        batch_size: int = 10,
        freq_save: int = 1,
        start_from: Optional[str] = None,
        can_run_in_parallel: bool = True,
    ) -> None:
        tps_init = time.time()

        sampler_ = MySampler(my_sampler_params, self.D_sampling, self.L, self.N)

        saver_ = MySaver(
            N=self.N,
            D=self.D,
            D_sampling=self.D_sampling,
            L=self.L,
            scaler=scaler,
            batch_size=batch_size,
            freq_save=freq_save,
            list_idx_sampling=self.list_idx_sampling,
        )

        run_optim_map = RunOptimMAP(self.path_data_csv_out, self.max_workers)
        run_optim_map.main(
            dict_posteriors,
            sampler_,
            saver_,
            scaler,
            N_MCMC,
            T_MC,
            path_raw=self.path_raw,
            start_from=start_from,
            freq_save=freq_save,
            can_run_in_parallel=can_run_in_parallel,
        )

        results_optim_map = ResultsExtractorOptimMAP(
            self.path_data_csv_out_optim_map,
            self.path_img,
            self.path_raw,
            N_MCMC,
            T_MC,
            T_BI,
            freq_save,
            self.max_workers,
        )
        for model_name, posterior in dict_posteriors.items():
            results_optim_map.main(
                posterior=posterior,
                model_name=model_name,
                scaler=scaler,
                #
                list_idx_sampling=self.list_idx_sampling,
                list_fixed_values=self.list_fixed_values,
                #
                estimator_plot=self.plots_estimator,
                Theta_true_scaled=self.Theta_true_scaled,
            )

        duration = time.time() - tps_init  # is seconds
        duration_str = time.strftime("%H:%M:%S", time.gmtime(duration))
        msg = "Simulation and analysis finished. Total duration : "
        msg += f"{duration_str} s\n"
        print(msg)

        list_model_names = list(dict_posteriors.keys())
        return list_model_names

    def inversion_mcmc(
        self,
        dict_posteriors: Dict[str, Posterior],
        scaler: MyScaler,
        my_sampler_params: MySamplerParams,
        N_MCMC: int,
        T_MC: int,
        T_BI: int,
        plot_1D_chains: bool,
        plot_2D_chains: bool,
        plot_ESS: bool,
        plot_comparisons_yspace: bool,
        #
        batch_size: int = 10,
        freq_save: int = 1,
        start_from: Optional[str] = None,
        #
        regu_spatial_N0: Union[int, float] = np.infty,
        regu_spatial_scale: Optional[float] = 1.0,
        regu_spatial_vmin: Optional[float] = 1e-8,
        regu_spatial_vmax: Optional[float] = 1e8,
        #
        y_valid: Optional[np.ndarray] = None,
        sigma_a_valid: Optional[np.ndarray] = None,
        omega_valid: Optional[np.ndarray] = None,
        sigma_m_valid: Optional[np.ndarray] = None,
        #
        can_run_in_parallel: bool = True,
        point_challenger: Dict = {},
        list_CI: List[int] = [68, 90, 95, 99],
    ) -> None:
        tps_init = time.time()

        sampler_ = MySampler(my_sampler_params, self.D_sampling, self.L, self.N)

        saver_ = MySaver(
            self.N,
            self.D,
            self.D_sampling,
            self.L,
            scaler,
            batch_size=batch_size,
            freq_save=freq_save,
            list_idx_sampling=self.list_idx_sampling,
        )

        run_mcmc = RunMCMC(self.path_data_csv_out, self.max_workers)
        run_mcmc.main(
            dict_posteriors,
            sampler_,
            saver_,
            scaler,
            N_MCMC,
            T_MC,
            T_BI=T_BI,
            path_raw=self.path_raw,
            path_csv_mle=self.path_data_csv_out_optim_mle,
            path_csv_map=self.path_data_csv_out_optim_map,
            start_from=start_from,
            freq_save=freq_save,
            #
            regu_spatial_N0=regu_spatial_N0,
            regu_spatial_scale=regu_spatial_scale,
            regu_spatial_vmin=regu_spatial_vmin,
            regu_spatial_vmax=regu_spatial_vmax,
            #
            can_run_in_parallel=can_run_in_parallel,
        )

        results_mcmc = ResultsExtractorMCMC(
            self.path_data_csv_out_mcmc,
            self.path_img,
            self.path_raw,
            N_MCMC,
            T_MC,
            T_BI,
            freq_save,
            self.max_workers,
        )
        for model_name, posterior in dict_posteriors.items():
            results_mcmc.main(
                posterior=posterior,
                model_name=model_name,
                scaler=scaler,
                list_names=self.list_names_plots,
                list_idx_sampling=self.list_idx_sampling,
                list_fixed_values=self.list_fixed_values,
                #
                plot_1D_chains=plot_1D_chains,
                plot_2D_chains=plot_2D_chains,
                plot_ESS=plot_ESS,
                plot_comparisons_yspace=plot_comparisons_yspace,
                #
                estimator_plot=self.plots_estimator,
                analyze_regularization_weight=np.isfinite(regu_spatial_N0),
                list_lines_fit=self.list_lines_fit,
                Theta_true_scaled=self.Theta_true_scaled,
                list_lines_valid=self.list_lines_valid,
                y_valid=y_valid,
                sigma_a_valid=sigma_a_valid,
                omega_valid=omega_valid,
                sigma_m_valid=sigma_m_valid,
                point_challenger=point_challenger,
                list_CI=list_CI,
            )

        duration = time.time() - tps_init  # is seconds
        duration_str = time.strftime("%H:%M:%S", time.gmtime(duration))
        msg = "Simulation and analysis finished. Total duration : "
        msg += f"{duration_str} s\n"
        print(msg)

        list_model_names = list(dict_posteriors.keys())
        return list_model_names



import os
from typing import List

import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from beetroots.inversion.results.utils.abstract_util import ResultsUtil
from beetroots.inversion.results.utils.mc import histograms


class ResultsRegularizationWeights(ResultsUtil):

    __slots__ = (
        "model_name",
        "list_names",
        "path_img",
        "path_data_csv_out_mcmc",
        "N_MCMC",
        "T_MC",
        "T_BI",
        "freq_save",
        "effective_T_BI",
        "effective_T_MC",
        "D",
    )

    def __init__(
        self,
        model_name: str,
        path_img: str,
        path_data_csv_out_mcmc: str,
        N_MCMC: int,
        T_MC: int,
        T_BI: int,
        freq_save: int,
        D: int,
        list_names: List[str],
    ):
        self.model_name = model_name
        self.list_names = list_names
        self.path_img = path_img
        self.path_data_csv_out_mcmc = path_data_csv_out_mcmc

        self.N_MCMC = N_MCMC
        self.T_MC = T_MC
        self.T_BI = T_BI
        self.freq_save = freq_save
        self.effective_T_BI = T_BI // freq_save
        self.effective_T_MC = T_MC // freq_save
        self.D = D

    def read_data(self, list_mcmc_folders: List[str]):
        list_tau = np.zeros((self.N_MCMC, self.effective_T_MC, self.D))
        for seed, mc_path in enumerate(list_mcmc_folders):
            with h5py.File(mc_path, "r") as f:
                # try:
                list_tau[seed] = np.array(f["list_tau"])
                # except:
                #     return np.empty((self.D,)), False

        return list_tau

    def create_folders(self):
        folder_path_inter = f"{self.path_img}/regularization_weights"
        folder_path = f"{folder_path_inter}/{self.model_name}"

        for path_ in [folder_path_inter, folder_path]:
            if not os.path.isdir(path_):
                os.mkdir(path_)

        return folder_path

    def estimate_regu_weight(self, list_mcmc_folders: List[str]) -> np.ndarray:
        for i, mc_path in enumerate(list_mcmc_folders):
            if i == 0:
                with h5py.File(mc_path, "r") as f:
                    list_tau = np.array(f["list_tau"][self.effective_T_BI :])

            else:
                with h5py.File(mc_path, "r") as f:
                    list_tau = np.concatenate(
                        [
                            list_tau,
                            np.array(
                                f["list_tau"][self.effective_T_BI :],
                            ),
                        ]
                    )
        estimated_regu_weights = list_tau.mean(0)
        return estimated_regu_weights

    def main(
        self,
        list_mcmc_folders: List[str],
    ) -> np.ndarray:
        folder_path = self.create_folders()
        list_tau = self.read_data(list_mcmc_folders)

        print("starting plots of regularization weights")

        for seed in range(self.N_MCMC):
            for d in range(self.D):
                list_tau_sd = list_tau[seed, :, d] * 1
                assert list_tau_sd.shape == (self.effective_T_MC,)

                list_tau_sd_no_BI = list_tau_sd[self.effective_T_BI :] * 1
                tau_MMSE = list_tau_sd_no_BI.mean(0)
                IC_2p5 = np.percentile(list_tau_sd_no_BI, q=2.5, axis=0)
                IC_97p5 = np.percentile(list_tau_sd_no_BI, q=97.5, axis=0)
                assert isinstance(tau_MMSE, float), tau_MMSE
                assert isinstance(IC_2p5, float), IC_2p5
                assert isinstance(IC_97p5, float), IC_97p5

                title = f"MC {self.list_names[d]} spatial regularization"
                title += " weight"
                histograms.plot_1D_chain(
                    list_tau_sd,
                    None,
                    d,
                    folder_path,
                    title,
                    lower_bounds_lin=1e-8 * np.ones((1,)),
                    upper_bounds_lin=1e8 * np.ones((1,)),
                    N_MCMC=self.N_MCMC,
                    T_MC=self.T_MC,
                    T_BI=self.T_BI,
                )

                title = "posterior distribution of spatial regularization"
                title += f" weight of {self.list_names[d]}"
                histograms.plot_1D_hist(
                    list_tau_sd_no_BI,
                    None,
                    d,
                    folder_path,
                    title=title,
                    lower_bounds_lin=1e-8 * np.ones((1,)),
                    upper_bounds_lin=1e8 * np.ones((1,)),
                    seed=seed,
                    estimator=tau_MMSE,
                    IC_low=IC_2p5,
                    IC_high=IC_97p5,
                )

        # altogether
        list_tau_flatter = list_tau.reshape(
            (self.N_MCMC * self.effective_T_MC, self.D),
        )
        list_tau_flatter_no_BI = list_tau[:, self.effective_T_BI :].reshape(
            (self.N_MCMC * (self.T_MC - self.T_BI) // self.freq_save, self.D)
        )

        tau_MMSE, _ = list_tau_sd_no_BI.mean(0)  # (D,)
        IC_2p5 = np.percentile(list_tau_sd_no_BI, q=2.5, axis=0)  # (D,)
        IC_97p5 = np.percentile(list_tau_sd_no_BI, q=97.5, axis=0)  # (D,)
        assert tau_MMSE.shape == (self.D,)
        assert IC_2p5.shape == (self.D,)
        assert IC_97p5.shape == (self.D,)

        plt.figure(figsize=(8, 6))
        plt.title("regularization weights sampling")
        for d, name in enumerate(self.list_names):
            plt.semilogy(list_tau_flatter[:, d], label=name)

        for seed in range(self.N_MCMC):
            if seed == 0:
                plt.axvline(
                    seed * self.effective_T_MC + self.effective_T_BI,
                    c="k",
                    ls="--",
                    label="T_BI",
                )
            elif seed == 1:
                plt.axvline(
                    seed * self.effective_T_MC,
                    c="k",
                    ls="-",
                    label="new MC",
                )
                plt.axvline(
                    seed * self.effective_T_MC + self.effective_T_BI,
                    c="k",
                    ls="--",
                )

            else:
                plt.axvline(seed * self.effective_T_MC, c="k", ls="-")
                plt.axvline(
                    seed * self.effective_T_MC + self.effective_T_BI,
                    c="k",
                    ls="--",
                )

        plt.grid()
        plt.legend()
        # plt.tight_layout()
        plt.savefig(
            f"{folder_path}/mc_regu_weights.PNG",
            bbox_inches="tight",
        )
        plt.close()

        for d in range(self.D):
            title = "posterior distribution of spatial regularization weight"
            title = f" of {self.list_names[d]}"
            histograms.plot_1D_hist(
                list_tau_flatter_no_BI[:, d],
                None,
                d,
                folder_path,
                title,
                self.lower_bounds_lin,
                self.upper_bounds_lin,
                None,
                tau_MMSE[d],
                IC_2p5[d],
                IC_97p5[d],
            )

        list_estimation_tau = []
        for d in range(self.D):
            dict_ = {
                "model_name": self.model_name,
                "d": d,
                "MMSE": tau_MMSE[d],
            }
            for q in [0.5, 2.5, 5, 95, 97.5, 99]:
                dict_[f"per_{q}"] = np.percentile(
                    list_tau_flatter_no_BI[:, d],
                    q=q,
                )
            list_estimation_tau.append(dict_)

        df_estimation_tau = pd.DataFrame(list_estimation_tau)
        path_file = f"{self.path_data_csv_out_mcmc}/estimation_tau.csv"
        df_estimation_tau.to_csv(
            path_file,
            mode="a",
            header=not (os.path.exists(path_file)),
        )
        print("plots of regularization weights done")
        return



import os
from typing import Dict, List, Optional, Sequence

import numpy as np
import torch
from nnbma.networks import NeuralNetwork
from torch.func import hessian, jacfwd, vmap

from beetroots.modelling.forward_maps.abstract_exp import ExpForwardMap


class NeuralNetworkApprox(ExpForwardMap):
    r"""Forward model based on a neural network. For every entry :math:`n \in [1, N]`

    .. math::

        f : \theta_n \in \mathbb{R}^D \mapsto f(\theta_n) \in \mathbb{R}^L

    where :math:`f(\theta_n)` can be written with a left composition with the exponential, i.e., :math:`f(\theta_n) = \exp \circ (\ln f) (\theta_n)`. Here the neural network corresponds to :math:`\ln f(\theta_n)`.

    The neural network needs to be defined with the ``nnbma`` package, presented in :cite:t:`paludNeuralNetworkbasedEmulation2023`.
    Here are the links towards the corresponding
    `GitHub repository <https://github.com/einigl/ism-model-nn-approximation>`_,
    `PyPi package <https://pypi.org/project/nnbma/>`_
    and `documentation <https://ism-model-nn-approximation.readthedocs.io/en/latest/?badge=latest>`_.
    """

    LOGE_10 = np.log(10.0)
    r"""float: natural log (in base :math:`e`) of 10, computed once and saved to limit redundant computations"""

    SIMU_PATH = os.path.abspath(f"{os.path.dirname(os.path.abspath(__file__))}/../../../data/")
    r"""str: path to the folder containing the data of saved models"""

    def __init__(
        self,
        model_name: str,
        dict_fixed_values_scaled: Dict[str, Optional[float]],
        device: Optional[str] = None,
    ):
        self.network = NeuralNetwork.load(
            model_name,
            f"{self.SIMU_PATH}/models/",
        )
        r"""NeuralNetwork: instance of neural network from the ``nnbma`` package (see https://pypi.org/project/nnbma/)"""

        assert device in [None, "cpu"]
        if device is None:
            # device = "cuda" if torch.cuda.is_available() else "cpu"
            device = "cpu" if device is None else device  # force cpu

        # set network to a specific device (either cpu or cuda)
        self.device = device
        r"""str: device on which the neural network is to be run, either cuda or cpu"""

        self.network.set_device(device)
        self.network.double().eval()

        self.D_no_kappa = self.network.input_features * 1
        r"""int: full dimension of the physical parameter vector, except for the scaling parameter :math:`\kappa` (which is not an input of the neural network)"""

        self.D = self.D_no_kappa + 1  # include kappa
        r"""int: full dimension of the physical parameter vector, including the scaling parameter :math:`\kappa`"""

        self.L = self.network.output_features * 1
        r"""int: total number of observables per pixel used for inversion"""

        self._update_derivatives()

        self.set_sampled_and_fixed_entries(dict_fixed_values_scaled)

        self.list_indices_to_sample_for_nn = [
            d - 1 for d in self.list_indices_to_sample if d >= 1
        ]  # remove kappa and start at Pth
        r"""list: indices of the entries to be sampled for the neural network, i.e., with an offset of 1 compared with `list_indices_fixed` due to the scaling parameter :math:`\kappa`"""

        self.list_indices_fixed_for_nn = [
            d - 1 for d in self.list_indices_fixed if d >= 1
        ]  # remove kappa and start at Pth
        r"""list: indices of the entries to be fixed for the neural network, i.e., with an offset of 1 compared with `list_indices_fixed` due to the scaling parameter :math:`\kappa`"""

        # display short message
        msg = f"neural network runs on : {self.network.device}"
        msg += f" (asked: {device})"
        print(msg)
        return

    def _update_derivatives(self):
        self.jacobian_network = vmap(jacfwd(self.network.forward))
        r"""function that yields the first derivative of the network"""
        self.hessian_network = vmap(jacfwd(jacfwd(self.network.forward)))
        r"""function that yields the diagonal of the second derivative of the network"""
        return

    def restrict_to_output_subset(self, output_subset: List[str]) -> None:
        """Restricts the full output set computed by the neural network to a potentially small subset of outputs. Permits to accelerate computations.

        Parameters
        ----------
        output_subset : List[str]
            list of the names of the outputs to be considered in forward map evaluations and derivatives.
        """
        for line in output_subset:
            assert line in self.network.current_output_subset, line

        self.network.restrict_to_output_subset(output_subset=output_subset)
        self.L = len(output_subset)

        self._update_derivatives()
        return

    def evaluate(self, Theta: np.ndarray) -> np.ndarray:
        return np.exp(self.evaluate_log(Theta))  # (N, L)

    def evaluate_log(self, Theta: np.ndarray) -> np.ndarray:
        Theta_combined = np.zeros((Theta.shape[0], self.D))
        Theta_combined += self.arr_fixed_values[None, :]
        for i, idx in enumerate(self.list_indices_to_sample):
            Theta_combined[:, idx] = Theta[:, i] * 1

        # the neural network returns log10 of intensities
        val = self.network.evaluate(Theta_combined[:, 1:])
        val *= self.LOGE_10  # go back to natural log scale

        val += Theta_combined[:, 0][:, None]  # add ln kappa

        return val  # (N, L)

    def gradient(self, Theta):
        msg = "Gradients should be computed with compute_all"
        raise NotImplementedError(msg)

    def gradient_log(self, Theta):
        msg = "Gradients should be computed with compute_all"
        raise NotImplementedError(msg)

    def _gradient_log(self, Theta):
        msg = "Gradients should be computed with compute_all"
        raise NotImplementedError(msg)

    def _hess_diag_log(self, Theta):
        msg = "Gradients should be computed with compute_all"
        raise NotImplementedError(msg)

    def hess_diag(self, Theta):
        msg = "Gradients should be computed with compute_all"
        raise NotImplementedError(msg)

    def hess_diag_log(self, Theta):
        msg = "Gradients should be computed with compute_all"
        raise NotImplementedError(msg)

    def _hess_full_log(self, Theta):
        msg = "Gradients should be computed with compute_all"
        raise NotImplementedError(msg)

    def compute_all(
        self,
        Theta: np.ndarray,
        compute_lin: bool = True,
        compute_log: bool = True,
        compute_derivatives: bool = True,
        compute_derivatives_2nd_order: bool = True,
    ) -> dict:
        r"""gathers the evaluation of the forward map in linear and log scale and of the associated derivatives. Permits to limit repeating computations, but requires the storage in memory of the result.

        Parameters
        ----------
        Theta : np.ndarray of shape (N, D)
            array of points in the input space :math:`\Theta = (\theta_n)_{n=1}^N` with :math:`\theta_n \in \mathbb{R}^D`
        compute_lin : bool, optional
            wether or not to compute the forward model (and possibly the gradient and diagonal of the Hessian), by default True
        compute_log : bool, optional
            wether or not to compute the log-forward model (and possibly the gradient and diagonal of the Hessian), by default True
        compute_derivatives : bool, optional
            wether or not to evaluate the derivatives of the forward map, by default True
        compute_derivatives_2nd_order : bool, optional
            wether or not to evaluate the 2nd order derivatives of the forward map, by default True


        Returns
        -------
        forward_map_evals : dict[str, np.ndarray]
            dictionary with entries such as `f_Theta`, `log_f_Theta`, `grad_f_Theta`, `grad_log_f_Theta`, `hess_diag_f_Theta` and `hess_diag_log_f_Theta`, depending on the input booleans.

        Note
        ----
        To evaluating :math:`f(\theta_n)` and the associated derivatives, 3 evaluations are enough for six functions. Calling each function would result in a total of 9 evaluations."""
        forward_map_evals = dict()

        N_pix = Theta.shape[0]

        # combine fixed and sampled values for model evaluation
        Theta_combined = np.zeros((N_pix, self.D))
        Theta_combined += self.arr_fixed_values[None, :]
        for i, idx in enumerate(self.list_indices_to_sample):
            Theta_combined[:, idx] += Theta[:, i]

        if compute_derivatives:
            _Theta = torch.from_numpy(Theta_combined[:, 1:])  # .float()

            #! integrated intensities and deriviatives in log10 scale
            log_f_Theta = self.network.forward(_Theta).detach().numpy()  # (N, L)

            assert log_f_Theta.shape == (
                N_pix,
                self.L,
            ), f"{log_f_Theta.shape} is not ({N_pix}, {self.L})"
            assert np.max(np.abs(log_f_Theta)) > 0

            grad_log_f_Theta = (
                self.jacobian_network(_Theta)
                .detach()
                .numpy()  # .to(self.network.device)
            ).transpose(
                (0, 2, 1)
            )  # (N, D, L)

            assert grad_log_f_Theta.shape == (
                N_pix,
                self.D_sampling,
                self.L,
            ), f"{grad_log_f_Theta.shape} is not ({N_pix}, {self.D_sampling}, {self.L})"
            assert np.max(np.abs(grad_log_f_Theta)) > 0

            log_f_Theta *= self.LOGE_10
            grad_log_f_Theta = grad_log_f_Theta[:, :, :] * self.LOGE_10

            if compute_derivatives_2nd_order:
                hess_full_log_f_Theta = (
                    self.hessian_network(_Theta)
                    .detach()
                    .numpy()  # .to(self.network.device)
                )  # (N, L, D, D)
                hess_diag_log_f_Theta = hess_full_log_f_Theta.diagonal(
                    offset=0, axis1=2, axis2=3
                ).transpose(
                    (0, 2, 1)
                )  # (N, D, L)

                assert hess_diag_log_f_Theta.shape == (
                    N_pix,
                    self.D_sampling,
                    self.L,
                ), f"{hess_diag_log_f_Theta.shape} is not ({N_pix}, {self.D_sampling}, {self.L})"
                assert np.max(np.abs(hess_diag_log_f_Theta)) > 0

                hess_diag_log_f_Theta = hess_diag_log_f_Theta[:, :, :] * self.LOGE_10

            # add log kappa
            log_f_Theta = Theta_combined[:, 0][:, None] + log_f_Theta

            if compute_log:
                forward_map_evals["log_f_Theta"] = log_f_Theta

                grad_log_f_Theta_full = np.ones((N_pix, self.D_sampling, self.L))
                if 0 in self.list_indices_to_sample:
                    grad_log_f_Theta_full[:, 1:, :] = (
                        grad_log_f_Theta[:, self.list_indices_to_sample[1:], :] * 1
                    )
                else:
                    grad_log_f_Theta_full[:, :, :] = (
                        grad_log_f_Theta[:, self.list_indices_to_sample, :] * 1
                    )
                forward_map_evals["grad_log_f_Theta"] = grad_log_f_Theta_full

                if compute_derivatives_2nd_order:
                    hess_diag_log_f_Theta_full = np.zeros(
                        (N_pix, self.D_sampling, self.L)
                    )
                    if 0 in self.list_indices_to_sample:
                        hess_diag_log_f_Theta_full[:, 1:, :] = (
                            hess_diag_log_f_Theta[:, self.list_indices_to_sample[1:], :]
                            * 1
                        )
                    else:
                        hess_diag_log_f_Theta_full[:, :, :] = (
                            hess_diag_log_f_Theta[:, self.list_indices_to_sample, :] * 1
                        )
                    forward_map_evals[
                        "hess_diag_log_f_Theta"
                    ] = hess_diag_log_f_Theta_full

            if compute_lin:
                f_Theta = np.exp(log_f_Theta)
                forward_map_evals["f_Theta"] = f_Theta

                # (N_pix, D, L)
                forward_map_evals["grad_f_Theta"] = (
                    grad_log_f_Theta_full * f_Theta[:, None, :]
                )

                # (N_pix, D, L)
                if compute_derivatives_2nd_order:
                    forward_map_evals["hess_diag_f_Theta"] = f_Theta[:, None, :] * (
                        hess_diag_log_f_Theta_full + grad_log_f_Theta_full**2
                    )

            return forward_map_evals

        else:
            log_f_Theta = self.evaluate_log(Theta)
            if compute_log:
                forward_map_evals["log_f_Theta"] = log_f_Theta
            if compute_lin:
                forward_map_evals["f_Theta"] = np.exp(log_f_Theta)
            return forward_map_evals



import os

import numpy as np
import pandas as pd

from beetroots.inversion.plots.map_shaper import MapShaper
from beetroots.modelling.forward_maps.neural_network_approx import NeuralNetworkApprox
from beetroots.simulations.astro.observation.abstract_observation import (
    SimulationObservation,
)
from beetroots.space_transform.transform import MyScaler


class SimulationToyCase(SimulationObservation):
    # def __init__(
    #     self,
    #     simu_name: str,
    #     cloud_name: str,
    #     N: int,
    #     dict_params_names: Dict[str, str],
    #     list_lines_fit: List[str],
    #     params: dict,
    #     max_workers: int = 10,
    #     small_size: int = 16,
    #     medium_size: int = 20,
    #     bigger_size: int = 24,
    # ):
    #     self.cloud_name = cloud_name
    #     # self.N = N
    #     # N_1_side = int(np.sqrt(self.N))

    #     self.list_lines_fit = list_lines_fit
    #     self.L = len(list_lines_fit)

    #     self.create_empty_output_folders(simu_name, params)
    #     self.setup_plot_text_sizes(small_size, medium_size, bigger_size)

    #     self.max_workers = max_workers

    #     self.list_names = list(dict_params_names.keys())
    #     self.list_names_plots = list(dict_params_names.values())
    #     # self.list_names_plots = [
    #     #     r"$\kappa$",
    #     #     r"$P_{th}$",
    #     #     r"$G_0$",
    #     #     r"$A_V^{tot}$",
    #     #     r"$\alpha$",
    #     # ]
    #     self.D = len(self.list_names)  # Number of physical parameters
    #     self.D_no_kappa = self.D - 1  # number of params that are used in nn

    def setup_observation(
        self,
        scaler: MyScaler,
        forward_map: NeuralNetworkApprox,
        sigma_a: np.ndarray,
        sigma_m: np.ndarray,
        omega: np.ndarray,
    ) -> pd.DataFrame:
        data_int_path = os.path.abspath(
            f"{os.path.dirname(os.path.abspath(__file__))}/../../../../data/toycases"
        )

        data_int_path += f"/{self.cloud_name}.pkl"
        syn_map = pd.read_pickle(data_int_path)

        for col in ["idx", "X", "Y"] + self.list_names:
            assert col in list(syn_map.columns), f"{col} not in {list(syn_map.columns)}"

        syn_map = syn_map.sort_values("idx")
        syn_map = syn_map.set_index(["X", "Y"])

        syn_map.loc[:, self.list_names] = scaler.from_lin_to_scaled(
            syn_map.loc[:, self.list_names].values
        )

        # * generate observations for specified lines
        self.Theta_true_scaled = syn_map.loc[:, self.list_names].values
        self.Theta_true_scaled = self.Theta_true_scaled[:, self.list_idx_sampling]

        syn_map.loc[:, self.list_lines_fit] = forward_map.evaluate(
            self.Theta_true_scaled,
        )

        # * save observations
        syn_map_to_csv = syn_map.copy()
        syn_map_to_csv.loc[:, self.list_names] = scaler.from_scaled_to_lin(
            syn_map_to_csv.loc[:, self.list_names].values,
        )
        syn_map_to_csv.to_csv(
            f"{self.path_data_csv_in}/true_params_and_emissions_maps.csv"
        )
        self.map_shaper = MapShaper(syn_map_to_csv)

        # * generation of noisy observation
        rng = np.random.default_rng(2023)
        eps_a = rng.normal(loc=0.0, scale=sigma_a)
        eps_m = rng.lognormal(
            mean=-(sigma_m**2) / 2,
            sigma=sigma_m,
        )
        assert eps_a.shape == (self.N, self.L), f"{(self.N, self.L)}"
        assert eps_m.shape == (self.N, self.L), f"{(self.N, self.L)}"
        f_Theta = forward_map.evaluate(self.Theta_true_scaled)
        y = np.maximum(omega, eps_m * f_Theta + eps_a)

        # rv_a = norm(loc=0, scale=sigma_a)
        # rv_m = lognorm(scale=np.exp(-(sigma_m ** 2) / 2), s=sigma_m)

        # y0 =forward_map.evaluate(self.Theta_true_scaled)

        # eps_a = rv_a.rvs(random_state=42)
        # eps_m = rv_m.rvs(random_state=42)
        # assert eps_a.shape == (self.N, self.L)
        # assert eps_m.shape == (self.N, self.L)

        # y = np.maximum(omega, eps_m * y0 + eps_a)  # (N,L)

        # * save observation
        df_observation = syn_map.copy()
        df_observation = df_observation.drop(columns=self.list_names)
        df_observation.loc[:, self.list_lines_fit] = y * 1
        df_observation.to_csv(f"{self.path_data_csv_in}/observation_maps.csv")

        return syn_map, y



from abc import ABC, abstractmethod
from typing import Union

import numpy as np


class PriorProbaDistribution(ABC):
    r"""Abstract Base Class for a probability distribution on non-countable set"""

    def __init__(self, D: int, N: int) -> None:
        self.D = D
        """int: number of distinct physical parameters"""
        self.N = N
        """int: number of pixels in each physical dimension"""

    @abstractmethod
    def neglog_pdf(self, Theta: np.ndarray) -> Union[float, np.ndarray]:
        pass

    @abstractmethod
    def gradient_neglog_pdf(self, Theta: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def hess_diag_neglog_pdf(self, Theta: np.ndarray) -> np.ndarray:
        pass
